// This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License.
// To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a
// letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.

//    Persistence of Vision Ray Tracer version 3.8 Include File.  (povr version ONLY)
//    File: functions.inc
//    Last updated: 16 July 2020
//    Description: This file contains the f_<name> function hooks
//    inbuilt and c++ compiled functions. Plus the definition for
//    F_<name> functions defined with the scene description language.
//    Many of the latter were formally in math.inc as 'sind' et cetera
//    looking like language keywords when they are not.


#ifndef(Functions_Inc_Temp)
#declare Functions_Inc_Temp = version;
#version 3.8; // For povr adding 'unoffical' makes some sense, but warnings then ugly.

#ifdef(View_POV_Include_Stack)
#   debug "including function.inc\n"
#end


#declare f_gradient = function { internal(0) }                  // was f_algbr_cyl1  (existed as 3 pattern wrapped X|Y|Z below)
// Parameters: x, y, z
    // Three extra parameter required:
    // 1. Gradient vector x value
    // 2. Gradient vector y value
    // 3. Gradient vector z value

#declare f_dents = function { internal(1) }                     // was f_algbr_cyl2  (existed as pattern wrapped below)
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Integer value indicating which underlying noise generator to use.
    //  Currently only values of 1-3 where 2 is POV-Ray's current default.
    //
    // Notes. This implementation like the old pattern wrapped one is NOT the
    // normal pattern implementation. The normal pattern implementation starts with
    // a f_dents result, but then multiplies three turbulence, a.k.a. DNoise(),
    // by it to get an always positive (abs) vector offset. The normal's bump
    // parameter controls how the turbulence offset is applied to the normal. In
    // other words, with the normal version it is possible to get inward or outward
    // turbulence based dents. To do this with functions for an isosurface will
    // take perhaps f_dents and three f_turb()s for the x,y,z offsets on the base
    // shape function call.

#declare f_radial = function { internal(2) }                    // was f_algbr_cyl3  (existed as pattern wrapped below)
// Parameters: x, y, z

#declare f_remap = function { internal(3) }                     // was f_algbr_cyl4
// Parameters:
    // 1. Incoming value to remap to weighted triangular peaked wave -1 to 1 to -1...
    // 2. Number of triangle peaks in values remap. Min of 2.
    // 3. Threshold to use for shell values... 0.0 returns weighted triangular values.
    //    When > 0.0, normalized negative value shells are created around the
    //    2-8 values specified with parameters [4-11].
    // 4-11 values to consider 0 points between peaks. Max of 8.

#declare f_cylinder = function { internal(4) }                 // Was f_bicorn
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Radius at y == 0.0.
    // Cylinder about the Y axis matching cylindrical pattern orientation.
    // Meaning x, z parameters are only ones used by default. Switch / swizzle
    // y with x or z to orient about unused axis.

#declare f_cone = function { internal(5) }                     // Was f_bifolia
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Radius.
    // Cone about the Y axis matching cylindrical pattern orientation
    // and really a dual cone with the points at abs(y) of +-1.0. To scale
    // the cone scale y. The y parameter divided by 2 is 2x taller, for example.

#declare f_granite23 = function { internal(6) }               // was f_blob
// Parameters: x, y, z
    // Six extra parameters used a trailing treatment switches taken in order.
    // 1. If != 0, Invert granite23 pattern result.
    // 2. If != 0, Apply a dents like pow(value,3) treatment. Old paper.
    // 3. If != 0, Invert current value
    // 4. If != 0, Apply cube root filter to exagerate range. Cool with 2.
    // 5. If != 0, Invert current value
    // 6. If != 0, Scaling of the value if other than 0.0.

#declare f_mult1to8pairs = function { internal(7) }           // was f_blob2
// Parameters:
    // 1. Number of value pairs.
    // 2. The specialized multiplication to use.
    //    If 0, Negative times negative is negative. All other
    //    polarity combinations result in a possitive value.
    //    If 1, Multiply pairs normally.
    //    If 2, Multiply pairs normally, sqrt of fabs, restore polarity.
    // 3. If 0, The returned value is the minimum of all pair products.
    //    If 1, The maximum of all. Add 10 to this return method and get
    //    the treatment, but instead a map index value returned.
    //    The index value returned will be the smallest (1-8 input) / 10.0.
    // 4-11. One to eight first in pair values.
    // 12-19. One to eight second in pair values.
    //
    // Note. The specialized multiplications 2(0) get a post multiply
    // square root for more accurate shapes. Without a small positive
    // bump like +0.005 ahead of isosurface use shapes look like harsher
    // f_max1to8pairs result at about 2x the cost.
    //
    // With the small positive bump the sharp seams often seen with max   <--- TODO. Should this be an additional parameter ???
    // are nicely softened AND the performance measure with the full
    // 8 pairs input is about 12.5% faster.
    //

#declare f_max1to8pairs = function { internal(8) }             // was f_boy_surface
// Parameters:
    // 1. Number of value pairs.
    // 2. Specialized methods. Currently unused.
    // 3. If 0, The returned value is the minimum of all pair maximums.
    //    If 1, The maximum of all. Add 10 to this return method and get
    //    the treatment but instead a map index value returned.
    //    The index value returned will be the smallest (1-8 input) / 10.0.
    // 4-11. One to eight first in pair values.
    // 12-19. One to eight second in pair values.

#declare f_min1to8pairs = function { internal(9) }             // was f_comma
// Parameters:
    // 1. Number of value pairs.
    // 2. Specialized methods. Currently unused.
    // 3. If 0, The returned value is the minimum of all pair minimums.
    //    If 1, The maximum of all. Add 10 to this return method and get
    //    the treatment but instead a map index value returned.
    //    The index value returned will be the smallest (1-8 input) / 10.0.
    // 4-11. One to eight first in pair values.
    // 12-19. One to eight second in pair values.

#declare f_morph2to9fncts = function { internal(10) }         // was f_cross_ellipsoids
// Parameters:
    // 1. Count of value inputs from 2 to 9. If set to 0, returns a map value
    //    where index value returned is the smallest (1-9 input) / 10.0
    //    and the associated weight is > 0.0.
    // 2. Clamp on how negative any given function contribution can be.
    // 3. Clamp on how positive any given function contribution can be.
    //    This one more important than (2) because letting any given function's
    //    positive, outside the shape, values get too large really limits what
    //    can be accomplished with respect to morphing. In fact, internal to
    //    f_morph2to9fncts(), the value specified is modified to be
    //    value/non-zero-weights.
    // 4-12. Input values (function calls) as parameters.
    // 13-21. Weights as portion of the sum of all weights.
    //
    // Notes. Sums and averages input values according to the weights.
    // Can morph in place or along a gradient. The math to control the
    // morphing - set the weights - done outside this function. The weights
    // themselves can be set up as 1/function-count or similar. Internally,
    // they get totaled and the particular functions contribution is its
    // weight against the total.
    //
    // A good starting place with the negative and positive clamp values
    // is -1 and 1 unless you know the max negative values with the torus
    // minor radius for example in which case start with +- that value.
    // If too many shapes, or too much of some shapes are disappearing
    // lower the positive clamp some more. You can of course set the clamps
    // to large values to avoid clamping altogether - most used when
    // morphing from one shape to another.
    //
    // The use of a prefilter in the code can cause false mx gradient warnings.
    //

#declare f_labsweep = function { internal(11) }               // Was f_crossed_trough
// Parameters: x, y, z
    // Seven extra parameter required:
    // 1. If 0, return new x.
    //    If 1, return new y.
    //    If 2, return new z.
    // 2. 3D point A x value.
    // 3. 3D point A y value.
    // 4. 3D point A x value.
    // 5. 3D point B x value.
    // 6. 3D point B y value.
    // 7. 3D point B x value.
    // Note. This does a linear sweep of a function about the origin
    // from point A to point B returning only the modified x, y
    // or z value. This enables us to code either directly on the function
    // about origin call, which can see modifiers. Or, we can code up each
    // as its own X,Y or X function to which we can apply pattern modifiers.
    // In testing the former simpler to code, less flexible and slower. The
    // latter, approach more difficult to code, more flexible and about 20%
    // faster when using the same modifiers for the same result.

#declare f_agate = function { internal(12) }                  // Was f_cubic_saddle
// Parameters: x, y, z
    // 7 extra parameter required:
    // 1. 0,1 or 2 for x, y or z agate alignment. Pattern's agate always uses z (2).
    // 2. Turbulence multiplier. Called agate_turb or inbuilt_turb_frequency in the
    //    pattern version. (1.0).
    // 3. Power exponent for pow(). Hard coded to 0.77 in the pattern version. (0.77)
    // 4. Number of octaves for the turbulence call. Pattern version defaults to (6).
    // 5. Omega value for the turbulence call. Pattern version defaults to 0.5.
    // 6. Lambda value for the turbulence call. Pattern version defaults to 2.0.
    // 7. The noise generator to use. Valid values 1-3. Pattern agate uses pattern or global setting.
    // Notes. To match the default pattern agate:
    // f_agate(x,y,z,2,1.0,0.77,6,0.5,2.0,2)  <-- Matches pattern agate when everything defaulted.
    // With the pattern agate octaves, omega and lambda can be set with those bare keywords
    // immediately after agate changing agate's inbuilt turbulence.

#declare f_onion = function { internal(13) }                  // Was f_cushion
// Parameters: x, y, z
    // Three extra parameter required:
    // 1. Onion layer depth. Very large values offer one layer, 0-? spherical gradient. Used
    //    with f_remap() for example to get onions with layers of varying thickness.
    // 2. Wave type. 0 = traditional ramp. 1 = triangular.
    // 3. If 1.0, returns values in a -1 to 1 range instead of traditinoal 0 to 1 range.

#declare f_sign = function { internal(14) }                   // Was f_devils_curve
// Parameters:
    // 1. Input value.
    // Notes. Returns 1 for >=0 and -1 for <0.

#declare f_sign0 = function { internal(15) }                  // Was f_devils_curve2d
// Parameters:
    // 1. Value
    // 2. Treament for 0. If 0, zero values return 0. If -1 zero values return -1.
    //    If 1, zero values return 1.
    // 3. Offset value. Identical to continuous pattern wave modifier 'phase.'
    // Notes. Excepting how 0 is treated given the parameter 2 setting, returns 1
    // for >0 and -1 for <0 after the parameter 3 offset is applied.

#declare f_cbrt = function { internal(16) }                   // Was f_dupin_cyclid
// Parameters:
    // 1. Input value.
    // 2. Number of additional cube roots.
    // Notes. Here making available the c++ standard library's cube root
    // function cbrt(). The expected main application is gradient reduction.
    // Unlike sqrt(), this function can handle negative input values returning
    // a negative cube root.

#declare f_ellipsoid = function { internal(17) }
// Parameters: x, y, z
    // Six extra parameters required:
    // 1. If 0, calculations are done with double floats. Otherwise with
    //    single floats. The latter often a lot faster with not much loss
    //    in accuracy.
    // 2. x scale
    // 3. y scale
    // 4. z scale
    // 5. pow() exponent to use. For a proper ellipsoid use 2.0. For an octahedron
    //    use 1.0. For an inward curvature from axises use values < 1.0. For
    //    Rounded box to round use values >2.0. The exponent here not locked
    //    at 2.0 offers some results like that of the superellipsoid at less
    //    compute expense.
    // 6. The number of times to take the sqrt() internally to reduce the gradient.
    //    1 a good starting point. Usually more than 2 not worth it. 0 for very
    //    round results works fine.
    // Notes. Previously f_ellipsoid() only worked for isosurfaces of threshold 1.0
    // and the axis scaling was inverted. In other words, it was not a general
    // function thought relatively fast. Changed in povr May 14, 2020 to align with
    // standard root/threshold behavior.  Further, extended to pass exponent as
    // this enables some f_superellipsoid results at a much lower compute cost.

#declare f_polyhedron = function { internal(18) }             // Was f_enneper
// Parameters: x, y, z as x, y z or passed f_gradient()s on arbitrary axises.
    // Thirteen additional parameters required:
    // 1. Count of input gradients. 3 to use only x,y,z. Max is 15.
    //    If the value is set negative a map value equal to the largest
    //    absolute input value divided by 16 is return. If first smallest, 1/16.
    // 2-13. Up to 12 input axis gradients. See f_gradient function, but
    //       any value input method allowed.
    // Notes. When using x,y,z or f_gradient() the gradients are projected
    // onto normalized input vectors. This is useful because it often allows
    // for some normalization of the scaling.
    //     #declare normScale = 3/(3+2*<additional_gradients>)
    //     f_polyhedron(x*normScale,y,z*normScale,5,
    //                  f_gradient(-1,0,1)*normScale,0,0,0,
    //                  0,0,0,0,
    //                  0,0,0,0)
    // Notes: When using the map values often one axis (y say) needs to be set to
    // 0.0 and other play might be required. Unsure how useful the map values will
    // be in practice...

#declare f_lame = function { internal(19) }                   // Was f_flange_cover
// Parameters:
    // 1. Number of input gradients. Up to eight.
    //    If the value is set negative a map value equal to the largest
    //    absolute input value divided by 10 is return. If first smallest, 1/10.
    // 2. If 0 does calculations as doubles, otherwise as floats for performance less accuracy.
    //    Here whether floats help depends a lot of the exponents. Mixed 0.6 to 1.6 on 5 gradients
    //    31.019 -> 28.882 ---> -6.89% going to single floats.
    // 3. The number of times to take the sqrt() internally to reduce the gradient.
    //    A value of 1 a good starting point. Usually more than 2 not worth it. 0 for very
    //    round results works fine.
    // 4-11. One to eight input gradient values.
    // 12-19. One to eight matching pow() exponent values.
    // Notes. If all exponents ==1, look at more efficient f_polyhedron alternative. Generally
    // if exponents are <1.0 introduces an inward / concave curvature. If 1, flat faces,
    // If >1 outward / convex curvature. See f_polyhedron comments on normalized scaling with
    // x,y,z projects onto normalized vectors.

#declare f_signpow = function { internal(20) }                // Was f_folium_surface
// Parameters:
    // 1. Input value.
    // 2. Exponent to apply to >= 0.0 values.
    // 3. Exponent to apply to < 0.0 values.
    // Notes. In all cases the pow function is run against the absolute value of the input. If
    // the input value was negative a negative result is returned. If positive, a positive result
    // is returned. Usually the exponent will be the same, but with the thought of helping with
    // gradients or complex implicit csg different exponents can be applied depending on whether
    // the input value is positive or negative.

#declare f_signsqrt = function { internal(21) }               // Was f_folium_surface2d
// Parameters:
    // 1. Input value.
    // Notes. The sqrt function is run against the absolute value of the input. The incoming
    // sign is applied to the result.

#declare f_hypot = function { internal(22) }                  // Was f_glob
// Parameters:
    // 1. First value (ie x)
    // 2. Second value (ie y)
    // Notes. Exposes the C++ standard hypot library function which encapsulates the common
    // sqrt(x*x + y*y) calculation.

#declare f_turbulence = function { internal(23) }             // Was f_heart
// Parameters: x, y, z
    // 8 extra parameter required:
    // 1. 0,1 or 2 for x, y or z result.
    // 2. The noise generator to use. Valid values 1-3.
    // 3. Scale returned internal turbulence() value for magnitude / (frequency for fmod inputs).
    //    Using 1.0 results in no scaling. This operation is the last performed before returning.
    // 4. Normalization.
    //    0 - None. Distribution will be centered at 0.0 with deviation depending upon turbulence
    //        settings of octaves, omega and lambda.
    //    1 - Adjust to 0 to 1, 0.5 centered distribution. Not clamped. If the passed
    //        deviation value not 'large enough' values can go outside the 0 to 1 range. See (3).
    //    2 - Adjust to -1 to +1, 0.0 centered distribution. Not clamped. See (2,4).
    //    3 - Basically (1) with hard clamping such that values below 0 are at 0 and values larger
    //        than 1 are at 1.
    //    4 - Basically (2) with hard clamping such that values below -1 are at -1 and values larger
    //        than +1 are at +1.
    //    5,7 - triangle, cycloidal to  0 to +1 range, respectively.
    //    6,8 - triangle, cycloidal to -1 to +1 range, respectively.
    // 5. Deviation. With default turbulence controls this will be 1 - and often less practically.
    //    This value should be the deviation as returned from turbulence(), ahead of any distribution
    //    adjustments. The value passed only used if (4) has a 1-4 value.
    // 6. Number of octaves for the turbulence call. Pattern version defaults to 6.
    // 7. Omega value for the turbulence call. Pattern version defaults to 0.5.
    // 8. Lambda value for the turbulence call. Pattern version defaults to 2.0.
    // Notes. (1) above provides for a method to get three substantially different values about any
    // given point in space. In other words, a way - for example - to get a vector turbulence with
    // three calls, one each for x,y,z.
    // Example: function { f_turbulence(x,y,z,0,2,1,0,1,6,0.5,2.0,2) }

#declare f_bump = function { internal(24) }                   // Was f_helical_torus
// Parameters:
    // 1. Input value as provided by a base function.
    // 2. Power of 2.0 exponent integer multiplier. For the typical rounded dome use 1.0. Higher
    //    values tend to square off or sharpen the result. Up to 20 usually OK.
    // 3. If 0, the value is returned straight up which means the maximum value is ~= 0.3679.
    //    If 1, return value is normalized to 0 to 1 range. If 2, the return value is normalized
    //    to a -1 to 0 range or a 0 to 1 range depending upon input polarity.
    // Notes. The bump function 'bumps' values > -1 and < 1. Other input values
    // are ignored.

#declare f_helix1 = function { internal(25) }
// Parameters: x, y, z
    // Seven extra parameters required:
    // 1. Number of helixes - e.g. 2 for a double helix
    // 2. Period - is related to the number of turns per unit length
    // 3. Minor radius
    // 4. Major radius
    // 5. Shape parameter. If this is greater than 1 then the tube becomes fatter in the y direction
    // 6. Cross section type. Between 0 and 3. 0-square, 1-circle, 2-diamond, 3-concave-diamond.
    // 7. Cross section rotation angle (degrees). E.g. if you choose a square
    //    cross section and rotate it by 45 degrees you get a diamond cross section.

#declare f_helix2 = function { internal(26) }
// Parameters: x, y, z
    // Seven extra parameters required:
    // 1. Period - is related to the number of turns per unit length
    // 2. Minor radius
    // 3. Major radius
    // 4. Cross section type. Between 0 and 3. 0-square, 1-circle, 2-diamond, 3-concave-diamond.
    // 5. Cross section rotation angle (degrees). E.g. if you choose a square
    //    cross section and rotate it by 45 degrees you get a diamond cross section.

#declare f_hex_x = function { internal(27) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Has no effect  ??????????????????????????????

#declare f_hex_y = function { internal(28) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Has no effect  ??????????????????????????????

#declare f_hetero_mf = function { internal(29) }
// Parameters: x, y, z
    // Six extra parameters required:
    // 1. H
    // 2. lacunarity
    // 3. octaves
    // 4. offset
    // 5. T
    // 6. noise

#declare f_hunt_surface = function { internal(30) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Field Strength

#declare f_hyperbolic_torus = function { internal(31) }
// Parameters: x, y, z
    // Three extra parameters required:
    // 1. Field strength
    // 2. Major radius: separation between the centers of the tubes at the closest point
    // 3. Minor radius: thickness of the tubes at the closest point

#declare f_isect_ellipsoids = function { internal(32) }
// Parameters: x, y, z
    // Four extra parameters required:
    // 1. eccentricity. When less than 1, the ellipsoids are oblate,
    //    when greater than 1 the ellipsoids are prolate,
    //    when zero the ellipsoids are spherical
    //   (and hence the whole surface is a sphere)
    // 2. inverse size. Increase this to decrease the size of the surface
    // 3. Diameter. Increase this to increase the size of the ellipsoids
    // 4. threshold. Setting this parameter to 1 and the threshold to zero
    //    has exactly the same effect as setting this parameter to zero
    //    and the threshold to -1

#declare f_kampyle_of_eudoxus = function { internal(33) }
// Parameters: x, y, z
    // Three extra parameters required:
    // 1. Field strength
    // 2. Dimple: When zero, the two dimples punch right through
    //    and meet at the center. Non-zero values give less dimpling
    // 3. Closeness: Higher values make the two planes become closer

#declare f_kampyle_of_eudoxus_2d = function { internal(34) }
// Parameters: x, y, z
    // Six extra parameters required:
    // 1. Field strength
    // 2. Dimple: When zero, the two dimples punch right through and meet at the center. Non-zero values give less dimpling
    // 3. loseness: Higher values make the two planes become closer
    // 4. sor switch
    // 5. sor offset
    // 6. sor angle

#declare f_klein_bottle = function { internal(35) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Field Strength

#declare f_kummer_surface_v1 = function { internal(36) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Field Strength

#declare f_kummer_surface_v2 = function { internal(37) }
// Parameters: x, y, z
    // Four extra parameters required:
    // 1. Field strength
    // 2. Rod width (negative): Setting this parameter
    //    to larger negative values increases the diameter of the rods
    // 3. Divergence (negative): Setting this number to -1
    //    causes the rods to become approximately cylindrical.
    //    Larger negative values cause the rods to become fatter further from the origin.
    //    Smaller negative numbers cause the rods to become narrower away from the origin,
    //    and have a finite length
    // 4. Influences the length of half of the rods.
    //    Changing the sign affects the other half of the rods.
    //    0 has no effect

#declare f_lemniscate_of_gerono = function { internal(38) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Field Strength

#declare f_lemniscate_of_gerono_2d = function { internal(39) }
// Parameters: x, y, z
    // Six extra parameters required:
    // 1. Field strength
    // 2. size: increasing this makes the 2d curve larger and less rounded
    // 3. width: increasing this makes the 2d curve fatter
    // 4. sor switch
    // 5. sor offset
    // 6. sor angle

#declare f_mesh1 = function { internal(40) }
// Parameters: x, y, z
    // Five extra Parameters required:
    // 1. Distance between neighboring threads in the x direction
    // 2. Distance between neighboring threads in the z direction
    // 3. Relative thickness in the x and z directions
    // 4. Amplitude of the weaving effect. Set to zero for a flat grid
    // 5. Relative thickness in the y direction

#declare f_mitre = function { internal(41) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Field Strength

#declare f_elliptical_cylinder = function { internal(42) }   // Was f_nodal_cubic()
// Parameters: x, y, z
    // Three extra parameter required:
    // 1. The traditional a multiplier of the x axis.
    // 2. The traditional b multiplier of the z axis.
    // 3. pow() exponent to use. For a true elliptical cylinder us 2.0. Flat
    //    sides use 1.0. For an inward curvature from axises use values < 1.0. For
    //    Rounded to round use values >2.0.

#declare f_elliptical_sphrswp = function { internal(43) }   // Was f_odd(), which was always a duplicate of f_cushion().
// Parameters: x, y, z
    // Six extra parameter required:
    // 1. The traditional a multiplier of the x axis.
    // 2. The traditional b multiplier of the z axis.
    // 3. The minor radius, r.
    // 4. The starting angle. 0 to 360.
    // 5. The arc angle from the start angle. -360 to 360.
    // 6. 0 = sweep with optimization. 1 = return 0 to 1 map value matched to angular rotation of arc.
    //    2 = skip up front in torus filter. (e.g. Using sweep values for map control)
    //
    // Notes. Like the f_torus() function the f_elliptical_sphrswp sits around
    // the y axis at the origin.  In other words, it is sitting in the x,z plane.
    // Angles and directional angles are left handed about y.
    //
    // With respect to isosurfaces; Gradients are decent, but due the internal
    // method of calculation, performance is somewhat volatile. Larger r values
    // tend to run more slowly. Rays more orthogonal to the y axis are slower -
    // sometimes slower than a parametric representation.
    //
    // The function is close to that of an elliptical torus - supertoroid with
    // e1,e2 at 1.0 (spherical) and asymmetric x,z axis values.  Also this method
    // has start and end points - arcs - there is less continuity / smoothness at
    // the cap spheres than elsewhere in each arc.


#declare f_ovals_of_cassini = function { internal(44) }
// Parameters: x, y, z
    // Four parameters required:
    // 1. Field Strength
    // 2. Major radius - like the major radius of a torus
    // 3. Filling. Set this to zero, and you get a torus. Set this to a higher
    //    value and the hole in the middle starts to heal up. Set it even higher and
    //    you get an ellipsoid with a dimple
    // 4. Thickness. The higher you set this value, the plumper is the result.

#declare f_paraboloid = function { internal(45) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Field Strength

#declare f_parabolic_torus = function { internal(46) }
// Parameters: x, y, z
    // Tree extra parameters required:
    // 1. Field Strength
    // 2. Major radius
    // 3. Minor radius

#declare f_ph = function { internal(47) }
// Parameters: x, y, z
    // no extra parameters required

#declare f_pillow = function { internal(48) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Field Strength

#declare f_piriform = function { internal(49) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Field Strength

#declare f_piriform_2d = function { internal(50) }
// Parameters: x, y, z
    // Seven extra parameters required:
    // 1. Field strength
    // 2. size factor 1: increasing this makes the curve larger
    // 3. size factor 2: making this less negative makes the curve larger but also thinner
    // 4. Fatness: increasing this makes the curve fatter
    // 5. sor switch
    // 6. sor offset
    // 7. sor angle

#declare f_poly4 = function { internal(51) }
// Parameters: x, y, z
    // Five extra parameters required:
    // 1. Constant
    // 2. y coefficient
    // 3. Y2 coefficient
    // 4. Y3 coefficient
    // 5. Y4 coefficient

#declare f_polytubes = function { internal(52) }
// Parameters: x, y, z
    // Six extra parameters required:
    // 1. Number of tubes
    // 2. Constant
    // 3. y coefficient
    // 4. Y2 coefficient
    // 5. Y3 coefficient
    // 6. Y4 coefficient

#declare f_quantum = function { internal(53) }
// Parameters: x, y, z

#declare f_quartic_paraboloid = function { internal(54) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Field Strength

#declare f_quartic_saddle = function { internal(55) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Field Strength

#declare f_quartic_cylinder = function { internal(56) }
// Parameters: x, y, z
    // Three extra parameters required:
    // 1. Field strength
    // 2. Diameter of the "egg"
    // 3. Controls the width of the tube and the vertical scale of the "egg"

#declare f_r = function { internal(57) }
// Parameters: x, y, z

#declare f_ridge = function { internal(58) }
// Parameters: x, y, z
    // Six extra parameters required:
    // 1.lambda
    // 2. octaves
    // 3. omega
    // 4. offset
    // 5. Ridge
    // 6. noise

#declare f_ridged_mf = function { internal(59) }
// Parameters: x, y, z
    // Six extra parameters required:
    // 1. H
    // 2. Lacunarity
    // 3. octaves
    // 4. offset
    // 5. Gain
    // 6. noise

#declare f_rounded_box = function { internal(60) }
// Parameters: x, y, z, P0, P1, P2, P3
    // Four extra parameters required:
    // 1. Radius of curvature.
    //    Zero gives square corners,
    //    0.1 gives corners that match "sphere {0,0.1}"
    // 2. scale x
    // 3. scale y
    // 4. scale z

#declare f_sphere = function { internal(61) }
// Parameters: x, y, z
    //One extra parameter required:
    // 1. radius of sphere

#declare f_spikes = function { internal(62) }
// Parameters: x, y, z
    // Five extra parameters required:
    // 1. Spikiness. Set this to very low values to increase the spikes.
    //    Set it to 1 and you get a sphere
    // 2. Hollowness. Increasing this causes the sides to bend in more
    // 3. size. Increasing this increases the size of the object
    // 4. Roundness. This parameter has a subtle effect on the roundness of the spikes
    // 5. Fatness. Increasing this makes the spikes fatter

#declare f_spikes_2d = function { internal(63) }
// Parameters: x, y, z
    // Four extra parameters required:
    // 1. Height of central spike
    // 2. frequency of spikes in the x direction
    // 3. frequency of spikes in the z direction
    // 4. Rate at which the spikes reduce as you move away from the center

#declare f_spiral = function { internal(64) }
// Parameters: x, y, z
    // Six extra parameters required:
    // 1. Distance between windings: setting this to 0.3 means that the spiral is 0.3 pov units further from the origin each time it completes one whole turn
    // 2. Thickness
    // 3. Outer diameter of the spiral. The surface behaves as if it is contained_by a sphere of this diameter
    // 4. Cross section type

#declare f_steiners_roman = function { internal(65) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Field Strength

#declare f_toupie = function { internal(66) }            // Was f_strophoid()
// Parameters: x, y, z
    // Four extra parameters required:
    // 1. If 0, uses doubles for calculations. If 1, uses floats. Floats less
    //    accurate, but faster.
    // 2. The major radius.
    // 3. Sort of the minor radius, but only truly that when the integer exponent(4)
    //    is 2.0 and a torus is the result. Otherwise, it is sort of a thickness in y parameter.
    // 4. Integer powers from 1 to 15. Values <1 set at 1. Values >15 mod 16. The greater
    //    this value, the flatter the top and bottom of the shape, though it's always ever
    //    slightly rounded even at large values. For even values, the larger values tend to fill
    //    in the middle/hole more so you more often get a button like result.
    //                               *
    //    f_toupie(x,y,z,0,0.2,0.005,1)  - Like a spindle torus intersection / lemon.
    //    f_toupie(x,y,z,0,0.5,0.005,2)  - Torus.
    //    f_toupie(x,y,z,0,0.5,0.005,3)  - Toupie. Odd exponents above all Toupie like.
    //    f_toupie(x,y,z,0,0.5,0.005,4)  - Torus nearly flat top/bottom. Metal washer.
    //    f_toupie(x,y,z,0,0.2,0.005,4)  - Torus nearly flat top/bottom. Solid middle like a button.
    //                                     Even values >4 all of similar form.
    //
    // Notes. This is only truly an f_toupie function when (4) is equal to 3. The internal
    // calculations are set up so the gradient is around 1.0 at (4) = 3 or the true f_toupie.
    // It's < 1.0 where (4) is < 3. Where (4) > 3 it creeps upward; when (4) = 6 the gradient
    // is around 2.5. At (4) = 12 the gradient sits around 11.5.


#declare f_supertorus = function { internal(67) }           // Was f_strophoid_2d()
// Parameters: x, y, z
    // Five extra parameters required:
    // 1. If 0, uses double floats to do the calculations. More accurate but often slower.
    //    If 1, uses single floats to do the calculations. See f_superellipsoid description
    //    for more information. Performance differences between float types are dramatic
    //    on many CPUs.
    // 2. Major radius.
    // 3. Minor radius.
    // 4. Square-ish, round, diamond-ish cross section >0, 1, >1 around supertorus y axis.
    // 5. Like 3 in some respects, but changes the supertorus shape itself.
    // Notes. When parameters 4 and 5 are near 1.0 the gradients are near 1.0 and the
    // supertorus looks like a torus. As either 4 or 5 approach 0 or 4 or so the gradients
    // can get quite high. 15 -> 80 say.

#declare f_superellipsoid = function { internal(68) }
// Parameters: x, y, z
// Five extra parameters required:
    // 1. If 0, uses double floats to do the calculations. On some hardware
    //    and recent c libraries, this can be slow. Slower than raw coded
    //    functions in the scene description language.
    //    If 1, uses single floats for the calculations which can be as
    //    much as 45% faster on some hardware. If working in singles enough
    //    to eliminate the bottleneck. The results ARE less accurate, but
    //    often OK.
    // 2. EW parameter. (doubles usable 0.01 to 4.0)
    // 3. NS parameter.
    // Note. At doubles EW/NS ranges of 0.01 to 4.0 workable. At single
    // floats ranges of 0.05 to 4.0 workable.
    //
    // NOTE! Do NOT take the EW or NS parameters to zero. The gradients
    // become absurd at or very near zero. Further, if one parameter very
    // small and the other large gradients can be >100. When both values
    // near 4.0, the gradients also start to push 100.
    //

#declare f_th = function { internal(69) }
// Parameters: x, y, z

#declare f_torus = function { internal(70) }
// Parameters: x, y, z
    // Two extra parameters required:
    // 1. Major radius
    // 2. Minor radius
    // Note. You can get the equivalent of the v38 spindle merge torus
    // by making the major radius smaller than the minor radius. Further
    // you an get the spindle intersection by keeping the magnitudes the
    // same as the merge torus but flipping the major radius value
    // negative -> f_torus(x,y,z,-0.1,0.3).
    // Aside: Where Major r > minor r sometimes called a ring torus.
    //        A major r <= minor r, merge, sometimes called a dorn torus.

#declare f_repeat = function { internal(71) }                 // was f_torus2
// Parameters: x, y, z
    // Thirteen extra parameter required:
    // 1.  If 0, return new x.
    //     If 1, return new y.
    //     If 2, return new z.
    //     If 3, return matched cells pattern equivalent value  0 to 1 interval.
    //     If 4, return matched cells pattern equivalent value -1 to 1 interval.
    // 2.  Count in x.
    // 3.  Count in y.
    // 4.  Count in z.
    // 5.  Delta in x.
    // 6.  Delta in y.
    // 7.  Delta in z.
    // 8.  Direction in x. If 0, count applied +-. If <0, count applied - only. If >0, count applied + only.
    // 9.  Direction in y.
    // 10. Direction in z.
    // 11. Cells pattern offset in x. Randomize to avoid duplicate cell values in local fmod() regions.
    // 12. Cells pattern offset in y.
    // 13. Cells pattern offset in z.
    // Notes. The base functino should be centered at zero and small enough to fit in the repeated 'cells'.
    // The counts too are in both directions about the particular axis. So a y count of two means two
    // on each side of y. The deltas represent the size of each cell in x,y or z. The directions are used
    // to prune counts in particular directions. Lastly, the cells offset exists so if you have multiple
    // instances of the isosurface you can make them all look different with respect to textures.

#declare f_torus_gumdrop = function { internal(72) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Field Strength

#declare f_umbrella = function { internal(73) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. Field Strength

#declare f_witch_of_agnesi = function { internal(74) }
// Parameters: x, y, z
    // Two extra parameters required:
    // 1. Field Strength
    // 2. Controls the width of the spike.
    //    The height of the spike is always about 1 unit

#declare f_witch_of_agnesi_2d = function { internal(75) }
// Parameters: x, y, z
    // Six extra parameters required:
    // 1. Field strength
    // 2. Controls the size of the spike
    // 3. Controls the height of the spike
    // 4. sor switch
    // 5. sor offset
    // 6. sor angle

#declare f_noise3d = function { internal(76) }
// Parameters: x, y, z
    // One extra parameter.
    // 1. The number of the noise generator to use. (1-3 currently) 2 - usual use.
    // Note. Returns values centered at 0.5 in a 0 to 1 range.

// NOTE! internal(77) is currently used for f_pattern. --> pattern {}

#declare f_snoise3d = function { internal(78) }
// Parameters: x, y, z
    // One extra parameter required:
    // 1. noise_generator number (1-3 currently) 2 - usual use.
    // Function returns noise in a -1 to 1 interval instead of the usual 0 - 1
    // with the interval centered at 0.

#declare f_npmod = function { internal(79) }
// Parameters: value, mod denominator, multiplier.
    // Implements the frequency + phase like mod function used with patterns.
    // Meaning the value goes to value/int(value/denominator) where the
    // value is not 0.0 at mod(value/demonimator) is 0.0.

#declare f_eblob = function { internal(80) }
// Parameters:  See external documentation for more detail.
    // 1. +- offset between two 1/exp() evaluations. 0.001 usually good.
    // 2. The number of active incoming functions / shapes from 1-8.
    // 3. The blobbing mode.
    // 4-11. The function value inputs. Unused are set to 0.0.
    // 12-19. The matching tightness values >0.0 to 100.0. Unused to 0.0.



// Isosurface pattern functions
// Some of these are not implemented because they require special
// parameters that must be specified in the definition. For this
// reason, you probably would want to define your own versions of
// these functions.
//--------------------
// f_agate           Replaced with inbuilt f_agate.
// average           Not implemented
#declare F_boxed     = function {pattern {boxed}} // was f_boxed
// f_bozo            Deleted. It was always only f_noise3d.
// f_brick           Deleted. Not commonly useful. pattern{} wrap user function if needed.
// f_bumps           Deleted. It was always only f_noise3d.
// f_cells           Never wrapped herein... Note! It is now available
//                   as modes(4,5) of the inbuilt f_repeat.
// f_checker         Deleted. Note commonly useful. pattern{} wrap user function if needed.
#declare F_crackle   = function {pattern {crackle}} // was f_crackle
// f_cylindrical     See f_cylinder otherwise {pattern {cylindrical}})
// density_file      Not implemented
// f_dents           Replaced with inbuilt f_dents.
// f_gradientX       Three replaced with single inbuilt f_gradient.
// f_gradientY
// f_gradientZ
// f_granite         Alternate inbuilt f_granite23 otherwise pattern{} wrap.
// image_pattern     Not implemented
// f_hexagon         Deleted. Not commonly useful. pattern{} wrap user function if needed.
#declare F_leopard   = function {pattern {leopard}} // was f_leopard
                     //Only the basic mandel pattern is implemented, its variants and
                     //The other fractal patterns are not implemented.
// f_mandel          Deleted. Never a good reason to wrap just one {pattern {mandel 25}}
#declare F_marble    = function {pattern {marble}} // was f_marble
// object            Not implemented
// f_onion           Replaced by inbuilt f_onion.
// pigment_pattern   Not implemented
#declare F_planar    = function {pattern {planar}} // was f_planar
// quilted           Not implemented
// f_radial          Replaced wtih inbuilt f_radial.
#declare F_ripples   = function {pattern {ripples}} // was f_ripples
// f_spherical       See f_sphere (r+fsphere()/r) or otherwise user {pattern {spherical}}
#declare F_spiral1   = function {pattern {spiral1 2}} // was f_spiral1
#declare F_spiral2   = function {pattern {spiral2 2}} // was f_spiral2
// f_spotted         Deleted. It was always only f_noise3d. Further, spotted pattern itself deleted.
#declare F_waves     = function {pattern {waves}} // was f_waves
#declare F_wood      = function {pattern {wood}} // was f_wood
#declare F_wrinkles  = function {pattern {wrinkles}} // was f_wrinkles


// Waveform functions
#declare F_sine_wave = function {sin(x*z*pi)*y}
    // f_sine_wave(val, amplitude, frequency)
#declare F_scallop_wave = function {abs(sin(x*z*pi)*y)}
    // f_scallop_wave(val, amplitude, frequency)

// ------------------------------------------------------------------
// Trig: (POV-Ray proper had these in math.inc without F_ prefix)
// ------------------------------------------------------------------

#declare F_sind   = function (x) {sin(radians(x))}
#declare F_cosd   = function (x) {cos(radians(x))}
#declare F_tand   = function (x) {tan(radians(x))}

#declare F_asind  = function (x)   {degrees(asin(x))}
#declare F_acosd  = function (x)   {degrees(acos(x))}
#declare F_atand  = function (x)   {degrees(atan(x))}
#declare F_atan2d = function (x,y) {degrees(atan2(x,y))}

// --------------------------------------------------------
// Misc: (POV-Ray proper had these in math.inc without F_ prefix)
// --------------------------------------------------------

#declare F_max3 = function (x,y,z) {max(x,y,z)}
#declare F_min3 = function (x,y,z) {min(x,y,z)}

#declare F_even = function(x) {select(mod(x,2),0,1,0)}
#declare F_odd  = function(x) {select(mod(x,2),1,0,1)}

// Squares the value
#declare F_sqr = function (x) {(x*x)}

// Returns the sign of a value. In povr use inbuilt f_sign instead.
#declare F_sgn = function (x) {select(x,-1,0,1)}

// Range handling clips a number (x) to the range [Min, Max] ([y, z]). Values
// above Max return Max, below Min return Min.
#declare F_clip = function (x,y,z) {min(z,max(x,y))}

// Clamps a number (x) to the range [Min, Max] ([y, z]).
// Values outside this range wrap around.
#declare F_clamp = function (x,y,z) {mod(x-y,z-y) + select(mod(x-y,z-y),z,y)}

// Adjusts input values (x) in the range [0, 1] to output values in range
// [Rmn,Rmx] ([y, z]). Normally (x) in input range, but it need not be.
#declare F_adj_range = function (x,y,z) {x*(z-y)+y}

// Adjusts passed (x) values relative to a specified range [Rmn, Rmx] to a new
// value relative the specified range [Min, Max]. Normally (x) in input range,
// but it need not be.
#declare F_adj_range2 = function (x,y,z,_mn,_mx) { ((x-y)/(z-y))*(_mx-_mn)+_mn }


#version Functions_Inc_Temp;
#end//functions.inc

