//******************************************************************************
///
/// @file vm/fnintern.cpp
///
/// This module implements built-in render-time functions.
///
/// This module is based on code by D. Skarda, T. Bily and R. Suzuki.
/// It includes functions based on code first introduced by many other
/// contributors.
///
/// @copyright
/// @parblock
///
/// Persistence of Vision Ray Tracer ('POV-Ray') version 3.8.
/// Copyright 1991-2020 Persistence of Vision Raytracer Pty. Ltd.
///
/// POV-Ray is free software: you can redistribute it and/or modify
/// it under the terms of the GNU Affero General Public License as
/// published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
///
/// POV-Ray is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Affero General Public License for more details.
///
/// You should have received a copy of the GNU Affero General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.
///
/// ----------------------------------------------------------------------------
///
/// POV-Ray is based on the popular DKB raytracer version 2.12.
/// DKBTrace was originally written by David K. Buck.
/// DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
///
/// @endparblock
///
//******************************************************************************

// Unit header file must be the first file included within POV-Ray *.cpp files (pulls in config)
#include "vm/fnintern.h"

// C++ variants of C standard header files
//  (none at the moment)

// C++ standard header files
#include <algorithm>

// POV-Ray header files (base module)
#include "base/mathutil.h"

// POV-Ray header files (core module)
#include "core/material/noise.h"
#include "core/material/pattern.h"
#include "core/material/pigment.h"
#include "core/material/warp.h"
#include "core/math/matrix.h"
#include "core/math/spline.h"
#include "core/scene/scenedata.h"
#include "core/scene/tracethreaddata.h"

// POV-Ray header files (VM module)
#include "vm/fnpovfpu.h"

// this must be the last file included
#include "base/povdebug.h"

namespace pov
{

using std::max;
using std::min;

static RandomDoubleSequence gPatternRands(0.0, 1.0, 32768);

static DBL fnHandle1to8ReturnCalc(const DBL ary[][4], const int valset, const int width, const int count,
        const int ReturnTreatment)
{
    int minValIndex = 0;
    int localReturnTreatment = ReturnTreatment & 0x07;
    bool returnMapIndexValue = (ReturnTreatment & 0x0A) ? true : false;

    DBL value = ary[0][valset];
    if (localReturnTreatment == 1)
    {
        for (int i=1; i<min(width,count); i++)
        {
            value = max(value,ary[i][valset]);
        }
    }
    else if (!returnMapIndexValue)  // (0) min of all the default.
    {
        for (int i=1; i<min(width,count); i++)
        {
            value = min(value,ary[i][valset]);
        }
    }

    if (returnMapIndexValue)
    {
        DBL tmpMinValue = ary[0][valset];
        for (int i=1; i<min(width,count); i++)
        {
            if (ary[i][valset] < tmpMinValue)
            {
                tmpMinValue = ary[i][valset];
                minValIndex = i;
            }
        }
        return (DBL)(minValIndex+1)/10.0;
    }
    else
        return value;
}

// Versions of follow two functions in textures.cpp too. Local static for performance.
static DBL fnCycloidal_Wave(DBL value)
{
    if (value >= 0.0)
        return sin((DBL) (((value - floor(value)) * 50000.0)) / 50000.0 * TWO_M_PI);
    else
        return 0.0-sin((DBL) (((0.0 - (value + floor(0.0 - value))) * 50000.0)) / 50000.0 * TWO_M_PI);
}

static DBL fnTriangle_Wave(DBL value)
{
    DBL offset;
    if (value >= 0.0)
        offset = value - floor(value);
    else
        offset = value + 1.0 + floor(fabs(value));
    if (offset >= 0.5)
        return (2.0 * (1.0 - offset));
    else
        return (2.0 * offset);
}

// Local preprocessor defines

/*
 * IMPORTANT NOTICE
 * Never ever use any PARAM_xxx macro after a call to POVFPU_SetLocal
 * because POVFPU_SetLocal may change the stack base address and thus
 * the "ptr" passed to the functions may be invalid after the call! [trf]
 */
#define PARAM_X (ptr[0])
#define PARAM_Y (ptr[1])
#define PARAM_Z (ptr[2])
#define PARAM(index) (ptr[index + 3])
#define PARAM_N_X(offset) (ptr[offset])
#define PARAM_N_Y(offset) (ptr[offset + 1])
#define PARAM_N_Z(offset) (ptr[offset + 2])
#define PARAM_N(index,offset) (ptr[index + offset + 3])

#define ROT2D(p,d,ang) if (p>0) {x2=sqrt(x2+PARAM_Z*PARAM_Z)- d; th=ang*M_PI_180; \
    if (th!=0){ PARAM_X= x2*cos(th)-PARAM_Y*sin(th); PARAM_Y= x2*sin(th)+PARAM_Y*cos(th);} else PARAM_X=x2; \
    x2=PARAM_X*PARAM_X; y2=PARAM_Y*PARAM_Y;}


// Global functions

DBL f_gradient(FPUContext *ctx, DBL *ptr, unsigned int fn); // 0
DBL f_dents(FPUContext *ctx, DBL *ptr, unsigned int fn); // 1
DBL f_radial(FPUContext *ctx, DBL *ptr, unsigned int fn); // 2
DBL f_remap(FPUContext *ctx, DBL *ptr, unsigned int fn); // 3
DBL f_cylinder(FPUContext *ctx, DBL *ptr, unsigned int fn); // 4
DBL f_cone(FPUContext *ctx, DBL *ptr, unsigned int fn); // 5
DBL f_granite23(FPUContext *ctx, DBL *ptr, unsigned int fn); // 6
DBL f_mult1to8pairs(FPUContext *ctx, DBL *ptr, unsigned int fn); // 7
DBL f_max1to8pairs(FPUContext *ctx, DBL *ptr, unsigned int fn); // 8
DBL f_min1to8pairs(FPUContext *ctx, DBL *ptr, unsigned int fn); // 9
DBL f_morph2to9fncts(FPUContext *ctx, DBL *ptr, unsigned int fn); // 10
DBL f_labsweep(FPUContext *ctx, DBL *ptr, unsigned int fn); // 11
DBL f_agate(FPUContext *ctx, DBL *ptr, unsigned int fn); // 12
DBL f_onion(FPUContext *ctx, DBL *ptr, unsigned int fn); // 13
DBL f_sign(FPUContext *ctx, DBL *ptr, unsigned int fn); // 14
DBL f_sign0(FPUContext *ctx, DBL *ptr, unsigned int fn); // 15
DBL f_cbrt(FPUContext *ctx, DBL *ptr, unsigned int fn); // 16
DBL f_ellipsoid(FPUContext *ctx, DBL *ptr, unsigned int fn); // 17
DBL f_polyhedron(FPUContext *ctx, DBL *ptr, unsigned int fn); // 18
DBL f_lame(FPUContext *ctx, DBL *ptr, unsigned int fn); // 19
DBL f_signpow(FPUContext *ctx, DBL *ptr, unsigned int fn); // 20
DBL f_signsqrt(FPUContext *ctx, DBL *ptr, unsigned int fn); // 21
DBL f_hypot(FPUContext *ctx, DBL *ptr, unsigned int fn); // 22
DBL f_turbulence(FPUContext *ctx, DBL *ptr, unsigned int fn); // 23
DBL f_bump(FPUContext *ctx, DBL *ptr, unsigned int fn); // 24
DBL f_helix1(FPUContext *ctx, DBL *ptr, unsigned int fn); // 25
DBL f_helix2(FPUContext *ctx, DBL *ptr, unsigned int fn); // 26
DBL f_hex_x(FPUContext *ctx, DBL *ptr, unsigned int fn); // 27
DBL f_hex_y(FPUContext *ctx, DBL *ptr, unsigned int fn); // 28
DBL f_hetero_mf(FPUContext *ctx, DBL *ptr, unsigned int fn); // 29
DBL f_hunt_surface(FPUContext *ctx, DBL *ptr, unsigned int fn); // 30
DBL f_hyperbolic_torus(FPUContext *ctx, DBL *ptr, unsigned int fn); // 31
DBL f_isect_ellipsoids(FPUContext *ctx, DBL *ptr, unsigned int fn); // 32
DBL f_kampyle_of_eudoxus(FPUContext *ctx, DBL *ptr, unsigned int fn); // 33
DBL f_kampyle_of_eudoxus_2d(FPUContext *ctx, DBL *ptr, unsigned int fn); // 34
DBL f_klein_bottle(FPUContext *ctx, DBL *ptr, unsigned int fn); // 35
DBL f_kummer_surface_v1(FPUContext *ctx, DBL *ptr, unsigned int fn); // 36
DBL f_kummer_surface_v2(FPUContext *ctx, DBL *ptr, unsigned int fn); // 37
DBL f_lemniscate_of_gerono(FPUContext *ctx, DBL *ptr, unsigned int fn); // 38
DBL f_lemniscate_of_gerono_2d(FPUContext *ctx, DBL *ptr, unsigned int fn); // 39
DBL f_mesh1(FPUContext *ctx, DBL *ptr, unsigned int fn); // 40
DBL f_mitre(FPUContext *ctx, DBL *ptr, unsigned int fn); // 41
DBL f_elliptical_cylinder(FPUContext *ctx, DBL *ptr, unsigned int fn); // 42
DBL f_elliptical_sphrswp(FPUContext *ctx, DBL *ptr, unsigned int fn); // 43
DBL f_ovals_of_cassini(FPUContext *ctx, DBL *ptr, unsigned int fn); // 44
DBL f_paraboloid(FPUContext *ctx, DBL *ptr, unsigned int fn); // 45
DBL f_parabolic_torus(FPUContext *ctx, DBL *ptr, unsigned int fn); // 46
DBL f_ph(FPUContext *ctx, DBL *ptr, unsigned int fn); // 47
DBL f_pillow(FPUContext *ctx, DBL *ptr, unsigned int fn); // 48
DBL f_piriform(FPUContext *ctx, DBL *ptr, unsigned int fn); // 49
DBL f_piriform_2d(FPUContext *ctx, DBL *ptr, unsigned int fn); // 50
DBL f_poly4(FPUContext *ctx, DBL *ptr, unsigned int fn); // 51
DBL f_polytubes(FPUContext *ctx, DBL *ptr, unsigned int fn); // 52
DBL f_quantum(FPUContext *ctx, DBL *ptr, unsigned int fn); // 53
DBL f_quartic_paraboloid(FPUContext *ctx, DBL *ptr, unsigned int fn); // 54
DBL f_quartic_saddle(FPUContext *ctx, DBL *ptr, unsigned int fn); // 55
DBL f_quartic_cylinder(FPUContext *ctx, DBL *ptr, unsigned int fn); // 56
DBL f_r(FPUContext *ctx, DBL *ptr, unsigned int fn); // 57
DBL f_ridge(FPUContext *ctx, DBL *ptr, unsigned int fn); // 58
DBL f_ridged_mf(FPUContext *ctx, DBL *ptr, unsigned int fn); // 59
DBL f_rounded_box(FPUContext *ctx, DBL *ptr, unsigned int fn); // 60
DBL f_sphere(FPUContext *ctx, DBL *ptr, unsigned int fn); // 61
DBL f_spikes(FPUContext *ctx, DBL *ptr, unsigned int fn); // 62
DBL f_spikes_2d(FPUContext *ctx, DBL *ptr, unsigned int fn); // 63
DBL f_spiral(FPUContext *ctx, DBL *ptr, unsigned int fn); // 64
DBL f_steiners_roman(FPUContext *ctx, DBL *ptr, unsigned int fn); // 65
DBL f_toupie(FPUContext *ctx, DBL *ptr, unsigned int fn); // 66
DBL f_supertorus(FPUContext *ctx, DBL *ptr, unsigned int fn); // 67
DBL f_superellipsoid(FPUContext *ctx, DBL *ptr, unsigned int fn); // 68
DBL f_th(FPUContext *ctx, DBL *ptr, unsigned int fn); // 69
DBL f_torus(FPUContext *ctx, DBL *ptr, unsigned int fn); // 70
DBL f_repeat(FPUContext *ctx, DBL *ptr, unsigned int fn); // 71
DBL f_torus_gumdrop(FPUContext *ctx, DBL *ptr, unsigned int fn); // 72
DBL f_umbrella(FPUContext *ctx, DBL *ptr, unsigned int fn); // 73
DBL f_witch_of_agnesi(FPUContext *ctx, DBL *ptr, unsigned int fn); // 74
DBL f_witch_of_agnesi_2d(FPUContext *ctx, DBL *ptr, unsigned int fn); // 75
DBL f_noise3d(FPUContext *ctx, DBL *ptr, unsigned int fn); // 76
DBL f_pattern(FPUContext *ctx, DBL *ptr, unsigned int fn); // 77
DBL f_snoise3d(FPUContext *ctx, DBL *ptr, unsigned int fn); // 78
DBL f_npmod(FPUContext *ctx, DBL *ptr, unsigned int fn); // 79
DBL f_eblob(FPUContext *ctx, DBL *ptr, unsigned int fn); // 80

void f_pigment(FPUContext *ctx, DBL *ptr, unsigned int fn, unsigned int sp); // 0
void f_transform(FPUContext *ctx, DBL *ptr, unsigned int fn, unsigned int sp); // 1
void f_spline(FPUContext *ctx, DBL *ptr, unsigned int fn, unsigned int sp); // 2
void f_patmods(FPUContext *ctx, DBL *ptr, unsigned int fn, unsigned int sp); // 3


// Global variables

constexpr Trap POVFPU_TrapTable[] =
{
    { f_gradient,                3 + 3 }, // 0
    { f_dents,                   1 + 3 }, // 1
    { f_radial,                  0 + 3 }, // 2
    { f_remap,                   8 + 3 }, // 3
    { f_cylinder,                1 + 3 }, // 4
    { f_cone,                    1 + 3 }, // 5
    { f_granite23,               6 + 3 }, // 6
    { f_mult1to8pairs,          16 + 3 }, // 7
    { f_max1to8pairs,           16 + 3 }, // 8
    { f_min1to8pairs,           16 + 3 }, // 9
    { f_morph2to9fncts,         18 + 3 }, // 10
    { f_labsweep,                7 + 3 }, // 11
    { f_agate,                   7 + 3 }, // 12
    { f_onion,                   3 + 3 }, // 13
    { f_sign,                    0 + 1 }, // 14
    { f_sign0,                   0 + 3 }, // 15
    { f_cbrt,                    0 + 2 }, // 16
    { f_ellipsoid,               6 + 3 }, // 17
    { f_polyhedron,             13 + 3 }, // 18
    { f_lame,                   16 + 3 }, // 19
    { f_signpow,                 0 + 3 }, // 20
    { f_signsqrt,                0 + 1 }, // 21
    { f_hypot,                   0 + 2 }, // 22
    { f_turbulence,              8 + 3 }, // 23
    { f_bump,                    0 + 3 }, // 24
    { f_helix1,                  7 + 3 }, // 25
    { f_helix2,                  5 + 3 }, // 26
    { f_hex_x,                   0 + 2 }, // 27
    { f_hex_y,                   0 + 2 }, // 28
    { f_hetero_mf,               6 + 3 }, // 29
    { f_hunt_surface,            1 + 3 }, // 30
    { f_hyperbolic_torus,        3 + 3 }, // 31
    { f_isect_ellipsoids,        4 + 3 }, // 32
    { f_kampyle_of_eudoxus,      3 + 3 }, // 33
    { f_kampyle_of_eudoxus_2d,   6 + 3 }, // 34
    { f_klein_bottle,            1 + 3 }, // 35
    { f_kummer_surface_v1,       1 + 3 }, // 36
    { f_kummer_surface_v2,       4 + 3 }, // 37
    { f_lemniscate_of_gerono,    1 + 3 }, // 38
    { f_lemniscate_of_gerono_2d, 6 + 3 }, // 39
    { f_mesh1,                   5 + 3 }, // 40
    { f_mitre,                   1 + 3 }, // 41
    { f_elliptical_cylinder,     3 + 3 }, // 42
    { f_elliptical_sphrswp,      6 + 3 }, // 43
    { f_ovals_of_cassini,        4 + 3 }, // 44
    { f_paraboloid,              1 + 3 }, // 45
    { f_parabolic_torus,         3 + 3 }, // 46
    { f_ph,                      0 + 3 }, // 47
    { f_pillow,                  1 + 3 }, // 48
    { f_piriform,                1 + 3 }, // 49
    { f_piriform_2d,             7 + 3 }, // 50
    { f_poly4,                   5 + 3 }, // 51
    { f_polytubes,               6 + 3 }, // 52
    { f_quantum,                 0 + 3 }, // 53
    { f_quartic_paraboloid,      1 + 3 }, // 54
    { f_quartic_saddle,          1 + 3 }, // 55
    { f_quartic_cylinder,        3 + 3 }, // 56
    { f_r,                       0 + 3 }, // 57
    { f_ridge,                   6 + 3 }, // 58
    { f_ridged_mf,               6 + 3 }, // 59
    { f_rounded_box,             4 + 3 }, // 60
    { f_sphere,                  1 + 3 }, // 61
    { f_spikes,                  5 + 3 }, // 62
    { f_spikes_2d,               4 + 3 }, // 63
    { f_spiral,                  4 + 3 }, // 64
    { f_steiners_roman,          1 + 3 }, // 65
    { f_toupie,                  4 + 3 }, // 66
    { f_supertorus,              5 + 3 }, // 67
    { f_superellipsoid,          3 + 3 }, // 68
    { f_th,                      0 + 3 }, // 69
    { f_torus,                   2 + 3 }, // 70
    { f_repeat,                 13 + 3 }, // 71
    { f_torus_gumdrop,           1 + 3 }, // 72
    { f_umbrella,                1 + 3 }, // 73
    { f_witch_of_agnesi,         2 + 3 }, // 74
    { f_witch_of_agnesi_2d,      6 + 3 }, // 75
    { f_noise3d,                 1 + 3 }, // 76
    { f_pattern,                 0 + 3 }, // 77
    { f_snoise3d,                1 + 3 }, // 78
    { f_npmod,                   0 + 3 }, // 79
    { f_eblob,                  16 + 3 }, // 80
    { nullptr, 0 }
};

constexpr TrapS POVFPU_TrapSTable[] =
{
    { f_pigment,                 0 + 3 }, // 0
    { f_transform,               0 + 3 }, // 1
    { f_spline,                  0 + 1 }, // 2
    { f_patmods,                 0 + 3 }, // 3
    { nullptr, 0 }
};

const unsigned int POVFPU_TrapTableSize = 81;
const unsigned int POVFPU_TrapSTableSize = 4;


// Functions

DBL f_gradient(FPUContext *ctx, DBL *ptr, unsigned int) // 0   was f_algbr_cyl1
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[0].parameter_cnt == 6u);
    DBL gx=PARAM(0), gy=PARAM(1), gz=PARAM(2);
    DBL value = dot(Vector3d(PARAM_X,PARAM_Y,PARAM_Z),Vector3d(gx,gy,gz).normalized());
    return (value);
}

DBL f_dents(FPUContext *ctx, DBL *ptr, unsigned int) // 1   was f_algbr_cyl2
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[1].parameter_cnt == 4u);
    int noiseGen = max(1,(int)PARAM(0) & 0x03);  // Currently 1-3 only allowed values.
    Vector3d Vec = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    DBL value =  Noise(Vec, noiseGen);
    value = value * value * value;
    return value;
}

DBL f_radial(FPUContext *ctx, DBL *ptr, unsigned int) // 2   was f_algbr_cyl3
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[2].parameter_cnt == 3u);
    DBL angle = atan2(PARAM_Z,PARAM_X);
    if (angle > 0.0)
    {
        angle = TWO_M_PI - angle;
    }
    DBL value = fabs(angle)/TWO_M_PI;
    return value;
}

DBL f_remap(FPUContext *ctx, DBL *ptr, unsigned int) // 3   was f_algbr_cyl4
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[3].parameter_cnt == 11u);
    const int width=8, height=3;
    DBL returnValue=1.0, ary[width][height];

    DBL value = PARAM_X;                     // Incoming value.
    int pCount = max(2,(int)PARAM_Y & 0x0F); // Number of 0 crossings.
    DBL thickness = PARAM_Z;                 // value thickness.

    int i,j;
    for (i=0; i<min(width,pCount); i++)
    {
        j = i - 1;
        ary[i][0] = PARAM(i);               // Additional triangle peaks.

        if (i > 0)
        {
            if (ary[i][0] <= ary[j][0])
            {
                 ary[i][0] = ary[j][0] + 1.0;  // Force increasing values.
            }

            ary[i][1] = ary[i][0] - (ary[i][0] - ary[j][0])/2.0;     // Triangle tip below this i threshold.
            ary[j][2] = ary[i][1];                                   // Triangle tip above prev j threshold.

            if (i == 1)
                ary[j][1] = ary[j][0] - (ary[i][0] - ary[j][0])/2.0; // Triangle tip below j=0 threshold.
        }
    }
    i -= 1;
    j = i - 1;
    ary[i][2] = ary[i][0] + (ary[i][0] - ary[j][0])/2.0; // Triangle tip above the last threshold.

    // If outside the ranges of interest return appropriate -1 or 1 value.
    if (value <= ary[0][1])
        return (thickness > 0.0) ? +1.0 : -1.0;
    if (value >= ary[pCount-1][2])
        return ((thickness > 0.0) || (pCount & 0x01)) ? +1.0 : -1.0;

    // To normalize the thickness must scan for maximum half delta.
    DBL minAboveBelowDelta = HUGE_VAL;
    if (thickness > 0.0)
    {
        for (i=0; i<min(width,pCount); i++)
        {
            DBL belowDelta = ary[i][0] - ary[i][1];
            DBL aboveDelta = ary[i][2] - ary[i][0];
            minAboveBelowDelta = min(minAboveBelowDelta,min(belowDelta,aboveDelta));
        }
    }
    //
    for (i=0; i<min(width,pCount); i++)
    {
        DBL fullDelta  = ary[i][2] - ary[i][1];
        DBL belowDelta = ary[i][0] - ary[i][1];
        DBL aboveDelta = ary[i][2] - ary[i][0];
        if ((value > ary[i][1]) && (value <= ary[i][2]))
        {
            DBL tmpVal = value - ary[i][1];
            if (tmpVal <= belowDelta)
            {
                if (i & 0x01)
                    returnValue = (0.5 - (0.5 * (tmpVal/belowDelta))) / +0.5;
                else
                    returnValue = (0.5 - (0.5 * (tmpVal/belowDelta))) / -0.5;
            }
            else
            {
                tmpVal = value - ary[i][0];
                if (i & 0x01)
                    returnValue = ((0.5 * (tmpVal/aboveDelta))) / -0.5;
                else
                    returnValue = ((0.5 * (tmpVal/aboveDelta))) / +0.5;
            }
            if (thickness > 0.0)
            {
                if (i & 0x01)
                {
                    returnValue = max(+returnValue-(thickness/2.0)*(minAboveBelowDelta/belowDelta),
                                      -returnValue-(thickness/2.0)*(minAboveBelowDelta/aboveDelta));
                }
                else
                {
                    returnValue = max(+returnValue-(thickness/2.0)*(minAboveBelowDelta/belowDelta),
                                      -returnValue-(thickness/2.0)*(minAboveBelowDelta/aboveDelta));
                }
            }
        }
    }
    return returnValue;
}

DBL f_cylinder(FPUContext *ctx, DBL *ptr, unsigned int) // 4   was f_bicorn
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[4].parameter_cnt == 4u);
    DBL value = std::hypot(PARAM_X,PARAM_Z) - PARAM(0);
    return value;
}

DBL f_cone(FPUContext *ctx, DBL *ptr, unsigned int) // 5   was f_bifolia
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[5].parameter_cnt == 4u);
    DBL value = std::hypot(PARAM_X,PARAM_Z) - PARAM(0) + fabs(PARAM_Y);
    return value;
}

DBL f_granite23(FPUContext *ctx, DBL *ptr, unsigned int) // 6   was f_blob
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[6].parameter_cnt == 9u);
    int i;
    DBL value = 0.0, freq = 1.0;
    Vector3d tv1, tv2;

    tv1 = Vector3d(PARAM_X,PARAM_Y,PARAM_Z) * 4.0;

    for (i = 0; i < 6; freq *= 2.0, i++)
    {
        tv2 = tv1 * freq;
        // In the block below tried the idea of using Perlin(3) and NoiseGen(2)
        // alternatively to better avoid artefacts. It looks good for artefacts
        // and coding up the alternative as f_granite23 and deleting f_granite

        if (i & 0x1)
        {
          //value += Noise(tv2,2) * (0.5/freq); // Fractional summed noise (wrinkle like?)
            value += min(0.5,fabs(1.0 - 2.0 * Noise(tv2,2))) * (0.5/freq);  // Granite hack, looks ok but a little different...
        }
        else
        {
          //value += Noise(tv2,3) * (0.50/freq);
            value += min(0.5,fabs(1.0 - 2.0 * Noise(tv2,3))) * (0.50/freq);
        }
    }

    value = (value <= 0.5) ? value : value + ((value-0.5)*0.70); // Granite hack? (YES and with inversion get puffy cloud look)

    // Comment these checks unless working on a better fit.
  //if (value > 1.0)
  //    throw POV_EXCEPTION_STRING("value > 1.0");
  //if (value < 0.0)
  //    throw POV_EXCEPTION_STRING("value < 0.0");

    value = max(0.0,min(1.0,value));

    if (PARAM(0) != 0.0)
        value = 1.0 - value;

    if (PARAM(1) != 0.0)
        value = value * value * value;  // A dents like treatment if passed unclamped sum...  (Use with cbrts below interesting...)

    if (PARAM(2) != 0.0)
        value = 1.0 - value;

    if (PARAM(3) != 0.0)                   // Amazing mountain range scale <1,5,1> at small gradient
        value = cbrt(cbrt(value) - value); // Really sharp peaks if also inverted...

    if (PARAM(4) != 0.0)
        value = 1.0 - value;

    if (PARAM(5) != 0.0)
        value *= PARAM(5);

    return value;
}

DBL f_mult1to8pairs(FPUContext *ctx, DBL *ptr, unsigned int) // 7   was f_blob2
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[7].parameter_cnt == 19u);
    const int width=8, height=4;
    DBL ary[width][height];
    int count = min(width,max(1,(int)PARAM_X & 0x0F)); // count of incoming pairs.
    int Method = (int)PARAM_Y & 0x0F;                  // Method of multiplication.
    int ReturnTreatment = (int)PARAM_Z & 0x0F;         // How to combine all the pair results. 0 = min. 1 = max.
                                                       // Add 10 to treatment to get map value back.

  //for (int i=0; i<min(width,count); i++) // In testing at 2 values this not faster. SIMD optimization below?
    for (int i=0; i<width; i++)
    {
        ary[i][0] = PARAM(i);       // incoming first in pair values.
        ary[i][1] = PARAM(width+i); // incoming second in pair values.

        if (Method == 0.0)
        {
            ary[i][2] = ((PARAM(i) < 0.0) && (PARAM(width+i) < 0.0)) ? -1.0 : 1.0;
        }
        else
        {
          //ary[i][2] = 1.0;
        }
    }

    if (Method == 0.0)
    {
      //for (int i=0; i<min(width,count); i++) // As above not faster...
        for (int i=0; i<width; i++)
        {
           ary[i][3] = sqrt(fabs(ary[i][0]) * fabs(ary[i][1])) * ary[i][2]; // sqrt controls values and gradients.
        }
    }
    else // Grouping other in else as 0 mode expected most common use.
    {
        if (Method == 2.0) // Traditional multiply followed by sqrt as with method 0, but after restoring polarity.
        {
            for (int i=0; i<width; i++)
            {
                ary[i][3] = ary[i][0] * ary[i][1];
                ary[i][2] = (ary[i][3] < 0.0) ? -1.0 : 1.0;
                ary[i][3] = sqrt(fabs(ary[i][3])) * ary[i][2];
            }
        }
        else
        {
            for (int i=0; i<width; i++)
            {
               ary[i][3] = ary[i][0] * ary[i][1];
            }
        }

        for (int i=0; i<width; i++)
        {
           ary[i][3] = ary[i][0] * ary[i][1];
        }
    }

    DBL value = fnHandle1to8ReturnCalc(ary,3,width,count,ReturnTreatment);
    return value;
}

// At the full 8 pairs, below 21% faster than 15 'vm' coded max()s.
DBL f_max1to8pairs(FPUContext *ctx, DBL *ptr, unsigned int) // 8   was f_max1to8pairs
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[8].parameter_cnt == 19u);
    const int width=8, height=4;
    DBL ary[width][height];
    int count = min(width,max(1,(int)PARAM_X & 0x0F)); // Count of incoming pairs.
    int Method = (int)PARAM_Y & 0x0F;                  // Method of maxes. Currently unused.
    int ReturnTreatment = (int)PARAM_Z & 0x0F;         // How to combine all the pair results. 0 = min. 1 = max.
                                                       // Add 10 to treatment to get map value back.

  //for (int i=0; i<min(width,count); i++) // In testing at 2 values this not faster. SIMD optimization below?
    for (int i=0; i<width; i++)
    {
        ary[i][0] = PARAM(i);       // incoming first in pair values.
        ary[i][1] = PARAM(width+i); // incoming second in pair values.
    }

    for (int i=0; i<min(width,count); i++)
    {
        ary[i][2] = max(ary[i][0],ary[i][1]);
    }

    DBL value = fnHandle1to8ReturnCalc(ary,2,width,count,ReturnTreatment);
    return value;
}

// At the full 8 pairs, below 21% faster than 15 'vm' coded min()s.
DBL f_min1to8pairs(FPUContext *ctx, DBL *ptr, unsigned int) // 9   was f_comma
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[9].parameter_cnt == 19u);
    const int width=8, height=4;
    DBL ary[width][height];
    int count = min(width,max(1,(int)PARAM_X & 0x0F)); // Count of incoming pairs.
    int Method = (int)PARAM_Y & 0x0F;                  // Method of mins. Currently unused.
    int ReturnTreatment = (int)PARAM_Z & 0x0F;         // How to combine all the pair results. 0 = min. 1 = max.
                                                       // Add 10 to treatment to get map value back.

  //for (int i=0; i<min(width,count); i++) // In testing at 2 values this not faster. SIMD optimization below?
    for (int i=0; i<width; i++)
    {
        ary[i][0] = PARAM(i);       // incoming first in pair values.
        ary[i][1] = PARAM(width+i); // incoming second in pair values.
    }

    for (int i=0; i<min(width,count); i++)
    {
        ary[i][2] = min(ary[i][0],ary[i][1]);
    }

    DBL value = fnHandle1to8ReturnCalc(ary,2,width,count,ReturnTreatment);
    return value;
}

DBL f_morph2to9fncts(FPUContext *ctx, DBL *ptr, unsigned int) // 10   was f_cross_ellipsoids
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[10].parameter_cnt == 21u);
    const int width=9, height=2;
    DBL ary[width][height];
    DBL valmin,valmax;
    int minValIndex;

    bool returnMapIndexValue = (PARAM_X == 0.0) ? true : false; // If count is zero -> map

    // First a prefilter to min of all values while doing some work for possible
    // map index return too. Note! The prefilter causes false max gradient warnings.
    valmax = max(+1e-6,PARAM_Z); // Positive value clamp.
    minValIndex = 0;
    valmin = PARAM(0);
    for (int i=1; i<width; i++)
        if (PARAM(i) < valmin)
        {
            if (PARAM(i+width) > 0.0)  // Ignore this value if weight 0.0
            {
                minValIndex = i;
                valmin = PARAM(i);
            }
        }
    if (returnMapIndexValue)
        return (DBL)(minValIndex+1)/10.0;
    else if (valmin > valmax)
        return valmin;

    // OK. Past prefilter and not returning map index so do all of it.
    int count = min(width,max(2,(int)PARAM_X & 0x0F)); // count of value, weight pairs.
    valmin = min(-1e-6,PARAM_Y); // Negative value clamp.

    for (int i=0; i<width; i++)
    {
        ary[i][0] = PARAM(i);                // incoming values.
        ary[i][1] = max(0.0,PARAM(width+i)); // incoming weights.
    }

    int positiveWeightCount = 0;
    DBL weightTotal = 0.0;
    for (int i=0; i<count; i++)
    {
        if (ary[i][1] > 0.0)
            positiveWeightCount += 1;
        weightTotal += ary[i][1];
    }
    if (weightTotal <= 0.0)
        return -1;
    for (int i=0; i<count; i++)
    {
        ary[i][1] /= weightTotal;  // normalize incoming weights.
    }

    valmax /= positiveWeightCount; // Done so positive values do not overwhelm negative shapes...
    DBL value = 0.0;
    for (int i=0; i<count; i++)
    {
        DBL tmpvalue = min(valmax,max(valmin,ary[i][0])); // Clamp to enable more morphing...
        value += (tmpvalue * (ary[i][1]));                // Sum values by normalized weights.
    }
    return value;
}

DBL f_labsweep(FPUContext *ctx, DBL *ptr, unsigned int) // 11   was f_crossed_trough
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[11].parameter_cnt == 10u);
    DBL value=0.0, x=PARAM_X, y=PARAM_Y, z=PARAM_Z;
    int xyOrz = (int)PARAM(0) & 0x03;  // 0=x, 1=y, 2=z.

    // Method here modeled on one posted by Tor Olav Kristensen which
    // implemented arbitrary point a to point b rounded cap cylinders.

    Vector3d pA = Vector3d(PARAM(1),PARAM(2),PARAM(3));
    Vector3d vL = Vector3d(PARAM(4),PARAM(5),PARAM(6)) - pA;
    DBL      LL = dot(vL,vL);
    Vector3d vD = vL/LL;
    DBL      AD = dot(pA,vD);

    if (xyOrz == 0)
        value = x - pA.x() - max(0.0,min(x*vD.x() + y*vD.y() + z*vD.z() - AD,1.0))*vL.x();
    else if (xyOrz == 1)
        value = y - pA.y() - max(0.0,min(x*vD.x() + y*vD.y() + z*vD.z() - AD,1.0))*vL.y();
    else if (xyOrz == 2)
        value = z - pA.z() - max(0.0,min(x*vD.x() + y*vD.y() + z*vD.z() - AD,1.0))*vL.z();

    return value;
}

DBL f_agate(FPUContext *ctx, DBL *ptr, unsigned int) // 12   was f_cubic_saddle
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[12].parameter_cnt == 10u);
    // f_agate(x,y,z,2,1.0,0.77,6,0.5,2.0,2)  <-- Matches pattern agate.
    DBL value       = 0.0;
    Vector3d EPoint = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    int xyOrz       = (int)PARAM(0) & 0x03;  // 0=x, 1=y, 2=z. (Note. pattern agate always uses z)
    DBL scaleTurb   = fabs(PARAM(1));               // pattern agate default is 1.0.
    DBL powExponent = fabs(PARAM(2));               // pattern agate always uses 0.77.

    TurbulenceWarp Turb;
    Turb.Octaves    = min(10,max(2,(int)PARAM(3) & 0x07)); // pattern agate defaults to 6;
    Turb.Omega      = PARAM(4);                            // pattern agate defaults to 0.5;
    Turb.Lambda     = min(10.0,max(-10.0,PARAM(5)));       // pattern agate degaults to 2.0;

    int noiseGen    = max(1,(int)PARAM(6) & 0x03);  // Currently 1-3 only allowed values.

    // Note. Agate default Turb->Turbulence vector values are 0.0 and that is OK.
    // Turbulence() uses only octaves, omega, lambda values.
    value    = scaleTurb * Turbulence(EPoint, &Turb, noiseGen);

    DBL xyOrzValue = PARAM_Z;
    if (xyOrz == 0)
        xyOrzValue = PARAM_X;
    else if (xyOrz == 1)
        xyOrzValue = PARAM_Y;

    value = 0.5 * (fnCycloidal_Wave(1.3 * value + 1.1 * xyOrzValue) + 1.0);

    value = max(0.0,min(1.0,value));

    value = pow(value, powExponent);

    return(value);
}

DBL f_onion(FPUContext *ctx, DBL *ptr, unsigned int) // 13   Was f_cushion
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[13].parameter_cnt == 6u);
    Vector3d EPoint = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    DBL layerDepth = max(gkMinIsectDepthReturned,fabs(PARAM(0)));  // Min 4.4e-8
    int wavetype = (int)PARAM(1) & 0x01;  // 0 ramp, 1 triangular
    bool functionInterval = (PARAM(2) != 1.0) ? false : true;

    DBL value;
    if ( wavetype == 0 )
    {
        value = fmod(EPoint.length(),layerDepth);
    }
    else
    {
        value = (fmod(EPoint.length(),layerDepth)-(layerDepth/2.0));
        value = fabs(value);
    }

    if ( functionInterval )
    {
        value = (2.0*value) - 1.0;
    }
    return value;
}

DBL f_sign(FPUContext *ctx, DBL *ptr, unsigned int) // 14   Was f_devils_curve
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[14].parameter_cnt == 1u);
    if (PARAM_X < 0.0)
        return -1.0;
    else
        return +1.0;
}

DBL f_sign0(FPUContext *ctx, DBL *ptr, unsigned int) // 15   Was f_devils_curve2d
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[15].parameter_cnt == 3u);
    DBL value = PARAM_X + PARAM_Z;
    if (value < 0.0)
        return -1.0;
    else if (value > 0.0)
        return +1.0;
    else
    {
        if (PARAM_Y == 0.0)
            return 0.0;
        else if (PARAM_Y < 0.0)
            return -1.0;
        else
            return +1.0;
    }
}

DBL f_cbrt(FPUContext *ctx, DBL *ptr, unsigned int) // 16   Was f_dupin_cyclid
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[16].parameter_cnt == 2u);
    DBL value = PARAM_X;
    int extras = (int)PARAM_Y & 0x0F; // Additional cbrt()s of input value.
    for (int i=0; i<=extras; i++)
    {
        value = std::cbrt(value);
    }
    return value;
}

// Previously f_ellipsoid() only worked for isosurfaces of threshold 1.0 and
// the axis scaling was inverted. In other words, it was not a general function
// thought relatively fast. Changed in povr May 14, 2020 to align with standard
// root/threshold behavior.  Further, extended to pass exponent as this enables
// some f_superellipsoid results at a much lower compute cost.
DBL f_ellipsoid(FPUContext *ctx, DBL *ptr, unsigned int) // 17
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[17].parameter_cnt == 9u);
    int sqrtCount = (int)PARAM(5) & 0x07; // How many times to take sqrt to reduce gradient.
    if (PARAM(0) == 0.0)
    {
        DBL x = PARAM_X,  y = PARAM_Y,  z = PARAM_Z;
        DBL a = PARAM(1), b = PARAM(2), c = PARAM(3);
        DBL e = PARAM(4);
        DBL value = pow((fabs(x/a)),e)+pow((fabs(y/b)),e)+pow((fabs(z/c)),e);
        value = sqrt(value);   // As with sphere form one sqrt() good initial.
        for (int i=0; i<sqrtCount; i++)
        {
            value = sqrt(value);   // Lowering gradient.
        }
        return value - 1.0;
    }
    else   // Use single floats. Often much faster at the expense of
           // a little accuarcy in practice. 3.46 -> 1.998s ---> -42.25%
    {
        float x = PARAM_X,  y = PARAM_Y,  z = PARAM_Z;
        float a = PARAM(1), b = PARAM(2), c = PARAM(3);
        float e = PARAM(4);
        float value = powf((fabs(x/a)),e)+powf((fabs(y/b)),e)+powf((fabs(z/c)),e);
        value = sqrtf(value);   // As with sphere form one sqrt() good initial.
        for (int i=0; i<sqrtCount; i++)
        {
            value = sqrtf(value);   // Lowering gradient.
        }
        return (DBL)(value -1.0);
    }

    // Original coding below. It is relatively quick.
    //return (sqrt(PARAM_X*PARAM_X * PARAM(0)*PARAM(0) +
    //             PARAM_Y*PARAM_Y * PARAM(1)*PARAM(1) +
    //             PARAM_Z*PARAM_Z * PARAM(2)*PARAM(2)));
}

// See f_lame for form with 'super' face curvature properties.
DBL f_polyhedron(FPUContext *ctx, DBL *ptr, unsigned int) // 18   Was f_enneper
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[18].parameter_cnt == 16u);
    const int width=4, height=4;
    DBL ary[width][height];
    bool returnMapIndexValue = (PARAM(0) < 0.0) ? true : false; // If count is less than zero -> map
    int count = min(width,max(1,(int)fabs(PARAM(0)) & 0x0F));   // count of incoming gradients / values.

    ary[0][0] = PARAM_X; ary[0][1] = PARAM_Y; ary[0][2] = PARAM_Z; ary[0][3] = 0.0;

  //for (int i=1; i<min(width,count); i++) // In testing at 2 values this not faster. SIMD optimization below?
    for (int i=0; i<width; i++)
    {
        ary[1][i] = (i<count)         ? PARAM(i+1)         : 0.0;  // incoming first set of 4 values / gradients.
        ary[2][i] = (width+i<count)   ? PARAM(width+i+1)   : 0.0;  // incoming second set of 4 values / gradients.
        ary[3][i] = (width+i+i<count) ? PARAM(width+i+i+1) : 0.0;  // incoming third set of 4 values / gradients.
    }

    int loops = count/width;

    DBL value = 0.0;
    int minValIndex = -1;
    if (returnMapIndexValue)
    {
        value = -HUGE_VAL;
        for (int i=0; i<=loops; i++)
            for (int j=0; j<4; j++)
                if ( ((fabs(ary[i][j])) > 0.0) && ((fabs(ary[i][j])) >= value) )
                    {
                        value = fabs(ary[i][j]);
                        minValIndex = (i * 4) + j;
                    }
        if (minValIndex<3)  // The ary[0][3] is skipped over to make code above simpler.
            return (DBL)(minValIndex+1)/16.0;
        else
            return (DBL)(minValIndex)/16.0;
    }
    else
    {
        for (int i=0; i<=loops; i++)
        {
            value += fabs(ary[i][0]) + fabs(ary[i][1]) + fabs(ary[i][2]) + fabs(ary[i][3]);
        }
    }

    value -= 1;
    return value;
}

// See f_polyhedron for form implementing e==1 flat faces more effeciently.
// SDL example:
// function { sqrt((pow((abs(x)/a),e)+pow((abs(y)/b),e)+pow((abs(z)/c),e) +
// pow((abs(f_gradient(x,y,z,-1,0,-1))/0.9),e) +
// pow((abs(f_gradient(x,y,z,-1,0,+1))/0.9),e)
// )*(3/7)) - 1.0}
//
DBL f_lame(FPUContext *ctx, DBL *ptr, unsigned int) // 19   Was f_flange_cover
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[19].parameter_cnt == 19u);
    const int width=8, height=2;
    bool returnMapIndexValue = (PARAM_X < 0.0) ? true : false; // If count is less than zero -> map
    int count = min(width,max(1,(int)fabs(PARAM_X) & 0x0F));   // count of incoming gradient, pow exponent pairs.
    int FloatMethod = (int)PARAM_Y & 0x0F;             // Whether using double (0) or single (!0) floats.
    int sqrtCount = (int)PARAM_Z & 0x07;               // How many times to take sqrt to reduce gradient.

    int minValIndex = -1;
    if (FloatMethod == 0)
    {
        DBL ary[width][height];
        DBL value = 0, maxVal = -HUGE_VAL;
      //for (int i=0; i<min(width,count); i++) // In testing at 2 values this not faster. SIMD optimization below?
        for (int i=0; i<width; i++)
        {
            ary[i][0] = PARAM(i);       // incoming first in pair values.
            ary[i][1] = PARAM(width+i); // incoming second in pair values.
        }
        for (int i=0; i<min(width,count); i++)
        {
            value += pow(fabs(ary[i][0]),ary[i][1]);
            if (returnMapIndexValue && (fabs(ary[i][0])>=maxVal))
            {
                maxVal = fabs(ary[i][0]);
                minValIndex = i;
            }
        }
        for (int i=0; i<sqrtCount; i++)
        {
            value = sqrt(value);
        }
        if (!returnMapIndexValue) return value -1.0;
    }
    else
    {
        float ary[width][height];
        float value = 0, maxVal = -HUGE_VAL;
      //for (int i=0; i<min(width,count); i++) // In testing at 2 values this not faster. SIMD optimization below?
        for (int i=0; i<width; i++)
        {
            ary[i][0] = PARAM(i);       // incoming first in pair values.
            ary[i][1] = PARAM(width+i); // incoming second in pair values.
        }
        for (int i=0; i<min(width,count); i++)
        {
            value += powf(fabs(ary[i][0]),ary[i][1]);
            if (returnMapIndexValue && (fabs(ary[i][0])>=maxVal))
            {
                maxVal = fabs(ary[i][0]);
                minValIndex = i;
            }
        }
        for (int i=0; i<sqrtCount; i++)
        {
            value = sqrtf(value);
        }
        if (!returnMapIndexValue) return (DBL)(value -1.0);
    }
    return (DBL)(minValIndex+1)/10.0; // Only reach when returnMapIndexValue true.
}

DBL f_signpow(FPUContext *ctx, DBL *ptr, unsigned int) // 20   Was f_folium_surface
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[20].parameter_cnt == 3u);
    DBL value = fabs(PARAM_X);
    if (value == 0.0) return 0.0;
    DBL sign  = (PARAM_X >= 0.0) ? +1.0 : -1.0;
    DBL ep    = max(PARAM_Y,gkMinIsectDepthReturned);
    DBL en    = max(PARAM_Z,gkMinIsectDepthReturned);
    if (sign == 1.0)
    {
        value = pow(value,ep);
    }
    else
    {
        value = pow(value,en);
    }
    return value*sign;
}

DBL f_signsqrt(FPUContext *ctx, DBL *ptr, unsigned int) // 21   Was f_folium_surface
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[21].parameter_cnt == 1u);
    DBL value = fabs(PARAM_X);
    if (value == 0.0) return 0.0;
    DBL sign  = (PARAM_X >= 0.0) ? +1.0 : -1.0;
    return sqrt(value)*sign;
}

DBL f_hypot(FPUContext *ctx, DBL *ptr, unsigned int) // 22   Was f_glob
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[22].parameter_cnt == 2u);
    return std::hypot(PARAM_X,PARAM_Y);
}

DBL f_turbulence(FPUContext *ctx, DBL *ptr, unsigned int) // 23    Was f_heart
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[23].parameter_cnt == 11u);
    DBL value       = 0.0;
    int xyOrz       = (int)PARAM(0) & 0x03;  // 0=x, 1=y, 2=z. (Use to get different x,y,z values at/near point)
    int noiseGen    = max(1,(int)PARAM(1) & 0x03);  // Currently 1-3 only allowed values.
    DBL freqTurb    = max(gkMinIsectDepthReturned,fabs(PARAM(2))); // scale magitude of turbulence() after any normalization.
    int normalize   = (int)PARAM(3) & 0x0F;  // 0 = none, 1 = 0 to +1, 2 = -1 to +1.
                                             //           3 clamps to, 4 clamps to range.
                                             //           5,7 triangle, cyloidal to  0 - +1 range.
                                             //           6,8 triangle, cyloidal to -1 - +1 range.
    DBL deviation   = fabs(PARAM(4));        // How far out is tail of distribution for normalizing opts 1 and 2.

    TurbulenceWarp Turb;
    Turb.Octaves    = min(10,max(2,(int)PARAM(5) & 0x0F)); // pattern turbulence defaults to 6;
    Turb.Omega      = PARAM(6);                            // pattern turbulence defaults to 0.5;
    Turb.Lambda     = min(10.0,max(-10.0,PARAM(7)));       // pattern turbulence defaults to 2.0;

    // By axis offsets necessary to avoid returning the same turbulence value
    // to each axis where function used at the same incoming x,y,z values to
    // realize an effective DTurbulence() (vector turbulence) without paying
    // 3x price of those calls for each return x,y or z offset.
    Vector3d EPoint;
    if (xyOrz == 0)
        EPoint = Vector3d(PARAM_X+(M_PI*sqrt(2)), PARAM_Y, PARAM_Z);
    else if (xyOrz == 1)
        EPoint = Vector3d(PARAM_X, PARAM_Y+(M_PI*sqrt(3)), PARAM_Z);
    else
        EPoint = Vector3d(PARAM_X, PARAM_Y, PARAM_Z+(M_PI*sqrt(5)));

    // Note. Default Turb->Turbulence vector values are 0.0 and that is OK.
    // Turbulence() uses only octaves, omega, lambda values.
    value = Turbulence(EPoint, &Turb, noiseGen);

    if (normalize > 0)
    {
        if (normalize > 4)
        {
            if (normalize == 5)
                value = fnTriangle_Wave(value);
            else if (normalize == 6)
                value = (fnTriangle_Wave(value) - 0.5) * 2.0;
            if (normalize == 7)
                value = (fnCycloidal_Wave(value) + 1.0) / 2.0;
            else if (normalize == 8)
                value = fnCycloidal_Wave(value);
        }
        else
        {
            //---
            if ((normalize == 1) || (normalize == 3))
            {
                value += deviation;       // Zero centered so shift distribution 'all' positive.
                value /= (deviation/0.5); // Value distribution to 0 to 1 range.
            }
            else if ((normalize == 2) || (normalize == 4))
                value /= (deviation/1.0); // Stretch to -1 to 1 range.
            //---
            if (normalize == 3)
                value = max(0.0,min(1.0,value));
            else if (normalize == 4)
                value = max(-1.0,min(1.0,value));
        }
    }

 // Vector3d result;
 // DTurbulence(result, EPoint, &Turb);  // !!! What about noise generator?  (At .x() looks OK, using DNoise)
 // DNoise(result, EPoint);              // This is centered and like 0 omega turbulence...
 // value = scaleTurb * result.x();
 // value = result.x();
 // value = Noise(EPoint,noiseGen);
//  void DTurbulence(Vector3d& result, const Vector3d& EPoint, const GenericTurbulenceWarp *Turb)

    return (value * freqTurb);
}

DBL f_bump(FPUContext *ctx, DBL *ptr, unsigned int) // 24   Was f_helical_torus
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[24].parameter_cnt == 3u);
    DBL value = max(-1.0+gkMinIsectDepthReturned,
                min(1.0-gkMinIsectDepthReturned,PARAM_X));
    DBL sign = (value < 0) ? -1.0 : +1.0;
    int powExpMult = (int)PARAM_Y & 0x7F;  // Usual rounded bump use multiplier of 1. 20 practical top depending.
    int normalize  = (int)PARAM_Z & 0x03;  // If 1, the result is normalized to 0-1 range.
                                           // If 2, result is normalized to -1.0 to 1.0 range with negative inputs
                                           // going toward -1 and positive toward 1.0 respectively. ???
    value = exp(1.0/(pow(value,2.0*max(1,powExpMult))-1));

    if (normalize <= 2)
    {
        value *= exp(1.0);
        if (normalize == 2)
            value *= sign;
    }

    return value;
}

DBL f_helix1(FPUContext *ctx, DBL *ptr, unsigned int) // 25
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[25].parameter_cnt == 10u);
    DBL r, r2, temp, th, ph, x2;
    /* helix                                     *
     *   p[0] : number of helix                  *
     *   p[1] : frequency                        *
     *   p[2] : minor radius                     *
     *   p[3] : major radius                     *
     *   p[4] : shape parameter                  *
     *   p[5] : cross section                    *
     *             p[5] = 1:  circle             *
     *                  = 2:  diamond            *
     *                  < 1:  rectangle(rounded) *
     *   p[6] : rotation angle for p[5]<1        */

    r = sqrt(PARAM_X * PARAM_X + PARAM_Z * PARAM_Z);
    if ((PARAM_X == 0) && (PARAM_Z == 0))
        PARAM_X = 0.000001;
    th = atan2(PARAM_Z, PARAM_X);
    th = fmod(th * PARAM(0) + PARAM_Y * PARAM(1) * PARAM(0), TWO_M_PI);
    if (th < 0)
        th += TWO_M_PI;
    PARAM_Z = (th - M_PI) / PARAM(4) / (PARAM(1) * PARAM(0));

    PARAM_X = r - PARAM(3);
    if (PARAM(5) == 1)
        r2 = sqrt(PARAM_X * PARAM_X + PARAM_Z * PARAM_Z);
    else
    {
        if (PARAM(6) != 0)
        {
            th = cos(PARAM(6) * M_PI_180);
            ph = sin(PARAM(6) * M_PI_180);
            x2 = PARAM_X * th - PARAM_Z * ph;
            PARAM_Z = PARAM_X * ph + PARAM_Z * th;
            PARAM_X = x2;
        }
        if (PARAM(5) != 0)
        {
            temp = 2. / PARAM(5);
            r2 = pow((pow(fabs(PARAM_X), temp) + pow(fabs(PARAM_Z), temp)), PARAM(5) *.5);
        }
        else
            r2 = max(fabs(PARAM_X), fabs(PARAM_Z));
    }

    return (-PARAM(2) + min((PARAM(3) + r), r2));
}

DBL f_helix2(FPUContext *ctx, DBL *ptr, unsigned int) // 26
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[26].parameter_cnt == 8u);
    DBL th, ph, x2, z2, r2, temp;
    // What does NFUNCTION mean?
    /* helical shape  for (minor radius>major radius  *
     *    cross section   p[4] same as NFUNCTION = 6      */
    th = PARAM_Y * PARAM(0);
    ph = cos(th);
    th = sin(th);
    x2 = PARAM_X - PARAM(2) * ph;
    z2 = PARAM_Z - PARAM(2) * th;
    PARAM_X = x2 * ph + z2 * th;
    PARAM_Z = (-x2 * th + z2 * ph);

    if (PARAM(4) == 1)
        return (sqrt(PARAM_X * PARAM_X + PARAM_Z * PARAM_Z) - PARAM(1));
    if (PARAM(4) != 0)
    {
        temp = 2. / PARAM(4);
        r2 = pow((pow(fabs(PARAM_X), temp) + pow(fabs(PARAM_Z), temp)), PARAM(4) *.5);
    }
    else
        r2 = max(fabs(PARAM_X), fabs(PARAM_Z));

    return (r2 - PARAM(1));
}

DBL f_hex_x(FPUContext *ctx, DBL *ptr, unsigned int) // 27
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[27].parameter_cnt == 2u);
    DBL x1,y1,x2,y2, th;
    x1=fabs(fmod(fabs(PARAM_X), sqrt(3.0))-sqrt(3.0)/2);
    y1=fabs(fmod(fabs(PARAM_Y), 3)-1.5);
    x2=sqrt(3.0)/2-x1;
    y2=1.5-y1;
    if ((x1*x1+y1*y1)>(x2*x2+y2*y2))
    {
        x1=x2;
        y1=y2;
    }
    if ((x1==0)&&(y1==0))
        PARAM_X=0.000001;
    th=atan2(y1,x1);
    if (th<M_PI/6)
        return(x1);
    else
    {
        x1=cos(M_PI/3)*x1+sin(M_PI/3)*y1;
        return(x1);
    }
}

DBL f_hex_y(FPUContext *ctx, DBL *ptr, unsigned int) // 28
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[28].parameter_cnt == 2u);
    DBL x1,y1,x2,y2, th;
    x1=fabs(fmod(fabs(PARAM_X), sqrt(3.0))-sqrt(3.0)/2);
    y1=fabs(fmod(fabs(PARAM_Y), 3)-1.5);
    x2=sqrt(3.0)/2-x1;
    y2=1.5-y1;
    if ((x1*x1+y1*y1)>(x2*x2+y2*y2))
    {
        x1=x2;
        y1=y2;
    }
    if ((x1==0)&&(y1==0))
        PARAM_X=0.000001;
    th=atan2(y1,x1);
    if (th<M_PI/6)
        return(y1);
    else
    {
        y1=-sin(M_PI/3)*x1+cos(M_PI/3)*y1;
        return(fabs(y1));
    }
}

DBL f_hetero_mf(FPUContext *ctx, DBL *ptr, unsigned int fn) // 29
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[29].parameter_cnt == 9u);
    DBL signal;
    Vector3d V1;
    DBL rem;

    V1 = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);

    int ngen = (int)PARAM(5) & 3;
    signal = (Noise(V1, ngen)*2.0 - 1.0) + PARAM(3);
    V1 *= PARAM(1);

    DBL p1_2_mp0 = pow(PARAM(1), -PARAM(0)), ea = p1_2_mp0;
    for (int i = 1; i < PARAM(2); i++)
    {
        // make a noisy increment and scale it by f^(-H)
        DBL inc = ((Noise(V1,ngen)*2.0 - 1.0) + PARAM(3)) * ea;

        // scale the increment by (the current signal)^PARAM(4) at V1
        //   so that PARAM(4)=0 gives 'straight' 1/f, PARAM(4)=1
        // gives heterogenous fractal, etc
        for (int p = (int) PARAM(4); p > 0; --p)
            inc *= signal;

        signal += inc;
        // go to next 'octave'
        V1 *= PARAM(1);
        ea *= p1_2_mp0;
    }

    rem = PARAM(2) - (int) PARAM(2);
    if(rem != 0.0)
    {
        // do something with fraction part of octave
        DBL inc = ((Noise(V1,ngen)*2.0-1.0) + PARAM(3)) * ea;
        signal += rem * inc * signal;
    }

    return signal;
}

DBL f_hunt_surface(FPUContext *ctx, DBL *ptr, unsigned int) // 30
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[30].parameter_cnt == 4u);
    DBL r, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, z2=PARAM_Z*PARAM_Z;
    r=-( 4*(x2+y2+z2-13)*(x2+y2+z2-13)*(x2+y2+z2-13) +
    27*(3*x2+y2-4*z2-12)*(3*x2+y2-4*z2-12)  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_hyperbolic_torus(FPUContext *ctx, DBL *ptr, unsigned int) // 31
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[31].parameter_cnt == 6u);
    DBL r, ph,th,x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, z2=PARAM_Z*PARAM_Z;
    ph=PARAM(1)*PARAM(1); th=PARAM(2)*PARAM(2);
    r=-( x2*x2 + 2*x2*y2 - 2*x2*z2 - 2*(ph+th)*x2 + y2*y2 -
    2*y2*z2 + 2*(ph-th)*y2 + z2*z2 + 2*(ph+th)*z2 + (ph-th)*(ph-th));
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_isect_ellipsoids(FPUContext *ctx, DBL *ptr, unsigned int) // 32
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[32].parameter_cnt == 7u);
    DBL r,x2,y2,z2;
    x2=PARAM_X*PARAM_X; y2=PARAM_Y*PARAM_Y; z2=PARAM_Z*PARAM_Z;
    /* f= min (f1, f2, f3) */
    r=min(exp(-(x2*PARAM(0)+y2*PARAM(0)+z2)*PARAM(1)), exp(-(x2*PARAM(0)+y2+z2*PARAM(0))*PARAM(1)));
    r=min(r, exp(-(x2+y2*PARAM(0)+z2*PARAM(0))*PARAM(1)));
    return(PARAM(3)-r*PARAM(2));
}

DBL f_kampyle_of_eudoxus(FPUContext *ctx, DBL *ptr, unsigned int) // 33
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[33].parameter_cnt == 6u);
    DBL r, x2=PARAM_X*PARAM_X;
    r=-( (PARAM_Y*PARAM_Y + PARAM_Z*PARAM_Z) - PARAM(2)*PARAM(2) * x2*x2 +  PARAM(2)*PARAM(2) * PARAM(1)*PARAM(1) * x2  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_kampyle_of_eudoxus_2d(FPUContext *ctx, DBL *ptr, unsigned int) // 34
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[34].parameter_cnt == 9u);
    DBL r, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, th;
    ROT2D(PARAM(3),PARAM(4),PARAM(5))
    r=-( y2  - PARAM(2)*PARAM(2) * x2*x2 +    PARAM(2)*PARAM(2) * PARAM(1)*PARAM(2) * x2  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_klein_bottle(FPUContext *ctx, DBL *ptr, unsigned int) // 35
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[35].parameter_cnt == 4u);
    DBL r, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, z2=PARAM_Z*PARAM_Z;
    r=-( (x2+y2+z2+2*PARAM_Y-1)*((x2+y2+z2-2*PARAM_Y-1)*(x2+y2+z2-2*PARAM_Y-1)-8*z2)+
    16*PARAM_X*PARAM_Z*(x2+y2+z2-2*PARAM_Y-1) );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_kummer_surface_v1(FPUContext *ctx, DBL *ptr, unsigned int) // 36
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[36].parameter_cnt == 4u);
    DBL r, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, z2=PARAM_Z*PARAM_Z;
    r=-( x2*x2+y2*y2+z2*z2-x2-y2-z2-x2*y2-x2*z2-y2*z2+1  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_kummer_surface_v2(FPUContext *ctx, DBL *ptr, unsigned int) // 37
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[37].parameter_cnt == 7u);
    DBL r, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, z2=PARAM_Z*PARAM_Z;
    r=-( x2*x2+y2*y2+z2*z2+PARAM(1)*(x2+y2+z2)+PARAM(2)*(x2*y2+x2*z2+y2*z2)+PARAM(3)*PARAM_X*PARAM_Y*PARAM_Z-1 );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_lemniscate_of_gerono(FPUContext *ctx, DBL *ptr, unsigned int) // 38
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[38].parameter_cnt == 4u);
    DBL r, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, z2=PARAM_Z*PARAM_Z;
    r=-(  x2*x2 - x2 + y2 + z2  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_lemniscate_of_gerono_2d(FPUContext *ctx, DBL *ptr, unsigned int) // 39
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[39].parameter_cnt == 9u);
    DBL r, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, th;
    ROT2D(PARAM(3),PARAM(4),PARAM(5))
    r=-( y2 - PARAM(2)*PARAM(2) * PARAM(1)*PARAM(1) * x2 +   PARAM(2)*PARAM(2)*x2*x2  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_mesh1(FPUContext *ctx, DBL *ptr, unsigned int) // 40
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[40].parameter_cnt == 8u);
    DBL th, ph, r, r2, temp;
    th = M_PI / PARAM(0);
    ph = M_PI / PARAM(1);
    r = fmod(PARAM_X, PARAM(0) * 2);
    if (r < 0)
        r += PARAM(0) * 2;
    r = fabs(r - PARAM(0)) * PARAM(2);
    r2 = (PARAM_Y - cos(PARAM_Z * ph) * PARAM(3)) * PARAM(4);
    temp = -sqrt(r2 * r2 + r * r);

    r = fmod(PARAM_X - PARAM(0), PARAM(0) * 2);
    if (r < 0)
        r += PARAM(0) * 2;
    r = fabs(r - PARAM(0)) * PARAM(2);
    r2 = (PARAM_Y + cos(PARAM_Z * ph) * PARAM(3)) * PARAM(4);
    temp = max(-sqrt(r2 * r2 + r * r), temp);

    r = fmod(PARAM_Z, PARAM(1) * 2);
    if (r < 0)
        r += PARAM(1) * 2;
    r = fabs(r - PARAM(1)) * PARAM(2);
    r2 = (PARAM_Y + cos(PARAM_X * th) * PARAM(3)) * PARAM(4);
    temp = max(-sqrt(r2 * r2 + r * r), temp);

    r = fmod(PARAM_Z - PARAM(1), PARAM(1) * 2);
    if (r < 0)
        r += PARAM(1) * 2;
    r = fabs(r - PARAM(1)) * PARAM(2);
    r2 = (PARAM_Y - cos(PARAM_X * th) * PARAM(3)) * PARAM(4);

    return (-max(-sqrt(r2 * r2 + r * r), temp));
}

DBL f_mitre(FPUContext *ctx, DBL *ptr, unsigned int) // 41
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[41].parameter_cnt == 4u);
    DBL r, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, z2=PARAM_Z*PARAM_Z;
    r=-( 4*x2*(x2 + y2 + z2) - y2*(1 - y2 - z2)  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_elliptical_cylinder(FPUContext *ctx, DBL *ptr, unsigned int) // 42   // Was f_nodal_cubic()
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[42].parameter_cnt == 6u);
    DBL x = PARAM_X,  y = PARAM_Y,  z = PARAM_Z;
    DBL a = PARAM(0), b = PARAM(1), e = max(gkDBL_epsilon,fabs(PARAM(2)));
    DBL value = pow((fabs(x/a)),e)+pow((fabs(z/b)),e);
    value = pow(value,min(1.0,1.0/e));   // As with sphere form one sqrt() good initial.
    return value - 1.0;
}

DBL f_elliptical_sphrswp(FPUContext *ctx, DBL *ptr, unsigned int) // 43   Was f_odd(), which was always a duplicate of f_cushion().
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[43].parameter_cnt == 9u);
    const int width=8, height=6; // Rows: 0|3 - Sphere sample vals; 2|4 - Rad at sample; 3|4 - three sequence value sums.
    DBL ary[width][height];

    DBL a = max(gkDBL_epsilon,fabs(PARAM(0))), b = max(gkDBL_epsilon,fabs(PARAM(1))), r = fabs(PARAM(2));

    int Method = (int)PARAM(5) & 0x03;  // 0 = sweep with optimization. 1 = return 0 to 1 map value.
                                        // 2 = skip up front torus filter. (e.g. Using sweep values for map control)
    DBL value;

    // If not inside slightly larger scaled torus, return quickly to avoid sampling.
    // Note! Metod 0 does create a significant discontinuity in the gradient, but with
    // respect to isosurfaces, this happens away from the surface and so doesn't
    // 'really' matter. The actual gradient near the surface below is always near 1.0.
    // TODO. Need smarter gradient reporting, some ability to turn off max
    // gradient warnings - OR both.
    if (Method != 2)
    {
        value = std::hypot(PARAM_X*b,PARAM_Z*a) - b*a;
        value = -(r+r/10) + std::hypot(value,PARAM_Y);
        if (value > 0.0)
            return value;
    }

    Vector3d Ept = Vector3d(PARAM_X,PARAM_Y,PARAM_Z);

    DBL degStart = min(360.0,max(0.0,fabs(PARAM(3))));  //
    DBL degOfArc = min(360.0,max(-360.0,PARAM(4)));     // Plus or minus arc OK.

    DBL radStart = -degStart * M_PI_180;
    DBL radOfArc = -degOfArc * M_PI_180;

    int steps = width - 1; // To get width value samples on each pass.
    DBL thisRadOfArcStep, thisRadOfArcHalfStep;

    auto lambdaEllipticalTorusSphereVal = []
         (const DBL xt, const DBL zt, const DBL r, const Vector3d ept) -> DBL
    {
        DBL sx = ept.x() - xt;
        DBL sz = ept.z() - zt;

        return sqrt(sx*sx + ept.y()*ept.y() + sz*sz)-r;
    };

    auto lambdaEllipticalSamples = [&ary, &thisRadOfArcStep, &thisRadOfArcHalfStep, &lambdaEllipticalTorusSphereVal]
         (const DBL thisRadStart, const DBL thisRadOfArc, const int steps,
          const DBL a, const DBL b, const DBL r, const Vector3d ept) -> void
    {
        thisRadOfArcStep     = thisRadOfArc/steps;
        thisRadOfArcHalfStep = thisRadOfArcStep / 2.0;
        DBL tmpSign = (thisRadOfArc > 0.0) ? +1.0 : -1.0;
        int scnt = 0;
        for (DBL radDeltaAngle = 0; radDeltaAngle <= tmpSign*(thisRadOfArc + thisRadOfArcHalfStep); radDeltaAngle += thisRadOfArcStep)
        {
            DBL thisAngle = thisRadStart + radDeltaAngle;
            ary[scnt][0] = lambdaEllipticalTorusSphereVal(a*cos(thisAngle),b*sin(thisAngle),r,ept);
            ary[scnt][1] = thisAngle;
            scnt += 1;
        }
    };

    if (radOfArc == 0.0)
    {
        value = lambdaEllipticalTorusSphereVal(a*cos(radStart),b*sin(radStart),r,Ept);
        return value;
    }

    int cnt;
    int cntMin, cntMin0;
    DBL thisRadStart[2],thisRadOfArc[2];
    thisRadStart[0] = radStart;
    thisRadOfArc[0] = thisRadOfArc[1] = radOfArc/2.0;
    thisRadStart[1] = radStart + thisRadOfArc[0];
    for (int h=0; h<2; h++)          // Handle arc in two halves to avoid numerical confusion in certain arcs >180 and <360. (h)alf.
    {
        for (int i=0; i<22; i++) // Limit at 22 for 'roughly' gkMinIsectDepthReturned accuracy. (Make a parameter? Can often be less)
        {
            lambdaEllipticalSamples(thisRadStart[h],thisRadOfArc[h],steps,a,b,r,Ept);
            if (fabs(thisRadOfArcStep) < gkMinIsectDepthReturned)
                break;

            //--- Calculate six, three value sum spans         //    01234567
            ary[0][2]  = ary[1][2] = (ary[1][0] +  ary[2][0]); // 12          1+2
            ary[0][2] += ary[0][0];                            // 0  012      12+0
            ary[1][2] += ary[3][0];                            // 1   123     12+3
            ary[2][2]  = ary[3][2] = (ary[3][0] +  ary[4][0]); // 34          3+4
            ary[2][2] += ary[2][0];                            // 2    234    34+2
            ary[3][2] += ary[5][0];                            // 3     345   34+5
            ary[4][2]  = ary[5][2] = (ary[5][0] +  ary[6][0]); // 56          5+6
            ary[4][2] += ary[4][0];                            // 4      456  56+4
            ary[5][2] += ary[7][0];                            // 5       567 56+7
          //ary[6][2]  = ary[7][2] = 0;
            //--- Find minimum and maximum 3 value sum.
            DBL maxVal = ary[0][2];
            DBL minVal = maxVal;
            for (cnt=1; cnt<width-2; cnt++)
            {
                maxVal = max(maxVal,ary[cnt][2]);
                minVal = min(minVal,ary[cnt][2]);
            }
            //--- Set cnt to maximum three value sum.
            for (cnt=0; cnt<width-2; cnt++)
                if (ary[cnt][2] == maxVal) break;
            //--- Set cntMin to minimum three value sum.
            for (cntMin=0; cntMin<width-2; cntMin++)
                if (ary[cntMin][2] == minVal) break;
            //--- Updates for next set of 8 samples.
            switch (cnt)
            {
                case 0:                         thisRadStart[h] = ary[2][1]; thisRadOfArc[h] = ary[7][1] - thisRadStart[h]; break;
                case 1:
                        {
                            if (cntMin < cnt) { thisRadStart[h] = ary[0][1]; thisRadOfArc[h] = ary[2][1] - thisRadStart[h]; }
                            else              { thisRadStart[h] = ary[2][1]; thisRadOfArc[h] = ary[7][1] - thisRadStart[h]; }
                            break;
                        }
                case 2:
                        {
                            if (cntMin < cnt) { thisRadStart[h] = ary[0][1]; thisRadOfArc[h] = ary[3][1] - thisRadStart[h]; }
                            else              { thisRadStart[h] = ary[3][1]; thisRadOfArc[h] = ary[7][1] - thisRadStart[h]; }
                            break;
                        }
                case 3:
                        {
                            if (cntMin < cnt) { thisRadStart[h] = ary[0][1]; thisRadOfArc[h] = ary[4][1] - thisRadStart[h]; }
                            else              { thisRadStart[h] = ary[4][1]; thisRadOfArc[h] = ary[7][1] - thisRadStart[h]; }
                            break;
                        }
                case 4:
                        {
                            if (cntMin < cnt) { thisRadStart[h] = ary[0][1]; thisRadOfArc[h] = ary[5][1] - thisRadStart[h]; }
                            else              { thisRadStart[h] = ary[5][1]; thisRadOfArc[h] = ary[7][1] - thisRadStart[h]; }
                            break;
                        }
                case 5:                         thisRadStart[h] = ary[0][1]; thisRadOfArc[h] = ary[5][1] - thisRadStart[h]; break;
                default: break;
            }
        }
        if (h == 0)
        {   // Remember results from first half arc.
            cntMin0 = cntMin;
            for (int j=0; j<height/2; j++)
                for (int k=0; k<width; k++)
                    ary[k][3+j] = ary[k][j];
        }
    }

    // Ending with 3 with thoughts of possible blobbing extensions lowering loop
    // depth - and ability to generate and orientation vector. For now min()s.
    if (Method == 1)
    {
        DBL valHalf0 = min(ary[cntMin0][3],min(ary[cntMin0+1][3],ary[cntMin0+2][3]));
        DBL valHalf1 = min(ary[cntMin][0],min(ary[cntMin+1][0],ary[cntMin+2][0]));
        if (valHalf0 <= valHalf1)
            value = max(0.0,min(1.0,fabs((ary[cntMin0+1][4] - radStart)/radOfArc)));
        else
            value = max(0.0,min(1.0,fabs((ary[cntMin+1][1] - radStart)/radOfArc)));
    }
    else
    {
        DBL valHalf0 = min(ary[cntMin0][3],min(ary[cntMin0+1][3],ary[cntMin0+2][3]));
        DBL valHalf1 = min(ary[cntMin][0],min(ary[cntMin+1][0],ary[cntMin+2][0]));
        value = min(valHalf0,valHalf1);
    }

    return value;
}

DBL f_ovals_of_cassini(FPUContext *ctx, DBL *ptr, unsigned int) // 44
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[44].parameter_cnt == 7u);
    DBL r, r2,x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, z2=PARAM_Z*PARAM_Z;
    r2 = (x2 + y2 + z2 + PARAM(1)*PARAM(1));
    r  = -(r2*r2 - PARAM(3)*PARAM(1)*PARAM(1)*(x2 + z2) - PARAM(2)*PARAM(2) );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_paraboloid(FPUContext *ctx, DBL *ptr, unsigned int) // 45
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[45].parameter_cnt == 4u);
    DBL r;
    r=-( PARAM_X*PARAM_X - PARAM_Y + PARAM_Z*PARAM_Z);
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_parabolic_torus(FPUContext *ctx, DBL *ptr, unsigned int) // 46
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[46].parameter_cnt == 6u);
    DBL r, ph,th, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, z2=PARAM_Z*PARAM_Z;
    ph=PARAM(1)*PARAM(1); th=PARAM(2)*PARAM(2);
    r=-(  x2*x2 + 2*x2*y2 - 2*x2*PARAM_Z - (ph+th)*x2 + y2*y2 - 2*y2*PARAM_Z +
    (ph-th)*y2 + z2 + (ph+th)*PARAM_Z +  (ph-th)* (ph-th)  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_ph(FPUContext *ctx, DBL *ptr, unsigned int) // 47
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[47].parameter_cnt == 3u);
    return( atan2(sqrt(PARAM_X*PARAM_X + PARAM_Z*PARAM_Z ),PARAM_Y) );
}

DBL f_pillow(FPUContext *ctx, DBL *ptr, unsigned int) // 48
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[48].parameter_cnt == 4u);
    DBL r, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, z2=PARAM_Z*PARAM_Z;
    r=( x2*x2 + y2*y2 + z2*z2 - (x2 + y2 + z2)  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_piriform(FPUContext *ctx, DBL *ptr, unsigned int) // 49
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[49].parameter_cnt == 4u);
    DBL r, x2=PARAM_X*PARAM_X;
    r=-( (x2*x2 - x2*PARAM_X) + PARAM_Y*PARAM_Y + PARAM_Z*PARAM_Z  );
    return( -min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_piriform_2d(FPUContext *ctx, DBL *ptr, unsigned int) // 50
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[50].parameter_cnt == 10u);
    DBL r, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y,th;
    ROT2D(PARAM(4),PARAM(5),PARAM(6))
    r=-( PARAM_Y*PARAM_Y - PARAM(1) * PARAM(3)* PARAM(3) * x2*PARAM_X -  PARAM(2) * PARAM(3)*PARAM(3) * x2*x2  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_poly4(FPUContext *ctx, DBL *ptr, unsigned int) // 51
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[51].parameter_cnt == 8u);
    // To get a Y cylinder of a radius specify PARAM(0) 'constant' value >0, with all other values at 0.0.
    // TODO. Might optimization be possible if coefficient inputs are sparse?
    DBL y2, temp;
    y2=PARAM_Y*PARAM_Y;
    temp=(PARAM(0)+PARAM(1)*PARAM_Y+PARAM(2)*y2 + PARAM(3)*y2*PARAM_Y +PARAM(4)*y2*y2);
    if (temp<-5.)
        temp=-5.;
    return (-temp+sqrt(PARAM_X*PARAM_X+PARAM_Z*PARAM_Z));
}

DBL f_polytubes(FPUContext *ctx, DBL *ptr, unsigned int) // 52
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[52].parameter_cnt == 9u);
    DBL x2,y2,z2,r,r2;
    int i;
    y2=PARAM_Y*PARAM_Y;
    r2=PARAM(1)+PARAM(2)*PARAM_Y+PARAM(3)*y2+PARAM(4)*y2*PARAM_Y+  PARAM(5)*y2*y2;
    r=-10000;
    for (i=0;i<(int)PARAM(0);i++)
    {
        x2 = PARAM_X - r2*sin(2*M_PI/PARAM(0)*i);
        z2 = PARAM_Z - r2*cos(2*M_PI/PARAM(0)*i);
        r= max(r,-sqrt( x2*x2 + z2*z2 ) );
    }
    return(-r);
}

DBL f_quantum(FPUContext *ctx, DBL *ptr, unsigned int) // 53
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[53].parameter_cnt == 3u);
    DBL r, th,temp;
    /*  well known function in quantum mechanics */
    if ((PARAM_X==0)&&(PARAM_Z==0))
        PARAM_X=1e-6;
    r=sqrt(PARAM_X*PARAM_X+PARAM_Z*PARAM_Z);
    th=cos(atan2(r,PARAM_Y));
    r=sqrt(r*r+PARAM_Y*PARAM_Y)*2.;
    temp=r*r*exp(-r*0.33333333)*(3.*th*th-1);
    return (temp*temp-12.0)*(-0.1);
}

DBL f_quartic_paraboloid(FPUContext *ctx, DBL *ptr, unsigned int) // 54
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[54].parameter_cnt == 4u);
    DBL r, x2=PARAM_X*PARAM_X, z2=PARAM_Z*PARAM_Z;
    r=-( x2*x2 + z2*z2 - PARAM_Y  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_quartic_saddle(FPUContext *ctx, DBL *ptr, unsigned int) // 55
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[55].parameter_cnt == 4u);
    DBL r, x2=PARAM_X*PARAM_X, z2=PARAM_Z*PARAM_Z;
    r=-( x2*x2 - z2*z2 - PARAM_Y  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_quartic_cylinder(FPUContext *ctx, DBL *ptr, unsigned int) // 56
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[56].parameter_cnt == 6u);
    DBL r, x2=PARAM_X*PARAM_X, z2=PARAM_Z*PARAM_Z;
    r=-( (x2 + z2) * PARAM_Y*PARAM_Y + PARAM(2)*PARAM(2) * (x2 + z2) - PARAM(2)*PARAM(2) * PARAM(1)*PARAM(1)  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_r(FPUContext *ctx, DBL *ptr, unsigned int) // 57
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[57].parameter_cnt == 3u);
    return( sqrt(PARAM_X*PARAM_X + PARAM_Y*PARAM_Y + PARAM_Z*PARAM_Z ) );
}

DBL f_ridge(FPUContext *ctx, DBL *ptr, unsigned int) // 58
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[58].parameter_cnt == 9u);
    Vector3d EPoint = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    int i;
    DBL Lambda, Omega, l, o;
    DBL value, v, resid, tot = 1.0, off, ridge, scale, rscale;
    Vector3d temp;
    int Octaves;
    int ngen = (int)PARAM(5) & 0x03;

    Lambda = l = PARAM(0);
    Octaves = (int)PARAM(1);
    Omega = o = PARAM(2);
    off = PARAM(3);
    ridge = PARAM(4);

    rscale = 1.0 / max(ridge, 1.0 - ridge);
    scale = 1.0 / max(off, 1.0 - off);
    resid = PARAM(1) - (DBL)Octaves;

    v = fabs(Noise(EPoint,ngen) - ridge) * rscale;
    value = (v - off);

    for (i = 2; i <= Octaves; i++)
    {
        temp = EPoint * l;
        v = fabs(Noise(temp,ngen) - ridge) * rscale;
        value += o * (v - off);
        tot += o;
        l *= Lambda;
        o *= Omega;
    }
    if (0.0 != resid)
    {
        temp = EPoint * l;
        v = fabs(Noise(temp,ngen) - ridge) * rscale;
        value += o * (v - off) * resid;
        tot += o * resid;
    }

    return (value * scale / tot);
}

DBL f_ridged_mf(FPUContext *ctx, DBL *ptr, unsigned int fn) // 59
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[59].parameter_cnt == 9u);
    FunctionCode *f = ctx->functionvm->GetFunction(fn);
    DBL *ea,freq,signal,weight,result;
    int i;
    Vector3d V1;
    int ngen = (int)PARAM(5) & 0x03;

    V1 = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    if (f->private_data == nullptr)
    {
        ea = reinterpret_cast<DBL *>(POV_MALLOC((PARAM(2) + 1)*sizeof(DBL), "exponent array"));
        freq = 1.0;
        for (i=0; i<=PARAM(2);i++)
        {
            ea[i]= pow(freq,-PARAM(0));
            freq *= PARAM(1);
        }
        f->private_data = reinterpret_cast<void *>(ea);
    }
    else
    {
        ea = reinterpret_cast<DBL *>(f->private_data);
    }
    signal = Noise(V1,ngen)*2.0-1.0;
    if (signal < 0.0 )
        signal = -signal;
    signal = PARAM(3) - signal;
    signal *= signal;
    result = signal;
    weight = 1.0;

    for (i=1; i<PARAM(2); i++)
    {
        V1 *= PARAM(1);
        weight = signal * PARAM(4);
        if (weight > 1.0)
            weight = 1.0;
        if (weight < 0.0)
            weight = 0.0;
        signal = Noise(V1,ngen)*2.0-1.0;
        if (signal < 0.0 )
            signal = -signal;
        signal = PARAM(3) - signal;
        signal *= signal;
        signal *= weight;
        result += signal * ea[i];
    }
    return (result);
}

DBL f_rounded_box(FPUContext *ctx, DBL *ptr, unsigned int) // 60
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[60].parameter_cnt == 7u);
    DBL x2, y2, z2, x3, y3, z3;

    x2 = PARAM(1) - PARAM(0);
    y2 = PARAM(2) - PARAM(0);
    z2 = PARAM(3) - PARAM(0);
    x3 = (PARAM_X < x2) ? 0 : (PARAM_X - x2);
    y3 = (PARAM_Y < y2) ? 0 : (PARAM_Y - y2);
    z3 = (PARAM_Z < z2) ? 0 : (PARAM_Z - z2);
    x2 = PARAM(0) - PARAM(1);
    y2 = PARAM(0) - PARAM(2);
    z2 = PARAM(0) - PARAM(3);
    PARAM_X = (PARAM_X > x2) ? 0 : (x2 - PARAM_X);
    PARAM_Y = (PARAM_Y > y2) ? 0 : (y2 - PARAM_Y);
    PARAM_Z = (PARAM_Z > z2) ? 0 : (z2 - PARAM_Z);
    PARAM_X = max(PARAM_X, x3);
    PARAM_Y = max(PARAM_Y, y3);
    PARAM_Z = max(PARAM_Z, z3);

    // Leave -1e-6 near as is. Isosurface use 1e-6 to 1e-8 at max solver accuracy.
    // Without the subtraction the 'not rounded corner' performance is extremely bad.
    return (-PARAM(0) + sqrt(PARAM_X * PARAM_X + PARAM_Y * PARAM_Y + PARAM_Z * PARAM_Z) - 1e-6);
}

DBL f_sphere(FPUContext *ctx, DBL *ptr, unsigned int) // 61
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[61].parameter_cnt == 4u);
  // Alternatives below without sqrt bad gradients lose big with current isosurface solver.
  //DBL value = (PARAM_X * PARAM_X + PARAM_Y * PARAM_Y + PARAM_Z * PARAM_Z - PARAM(0) * PARAM(0));
  //return min(1.0,max(-1.0,value));                                                              // 145.06 -> 295.24s ---> +103.53%
  //return (PARAM_X * PARAM_X + PARAM_Y * PARAM_Y + PARAM_Z * PARAM_Z - PARAM(0) * PARAM(0));     // 145.06 -> 288.72s --->  +99.03%
    return (-PARAM(0) + sqrt(PARAM_X * PARAM_X + PARAM_Y * PARAM_Y + PARAM_Z * PARAM_Z));
}

DBL f_spikes(FPUContext *ctx, DBL *ptr, unsigned int) // 62
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[62].parameter_cnt == 8u);
    DBL r,x2,y2,z2;
    x2=PARAM_X*PARAM_X; y2=PARAM_Y*PARAM_Y; z2=PARAM_Z*PARAM_Z;
    r= exp(-(x2+y2+z2)*PARAM(3))*PARAM(4)          -exp(-(x2*PARAM(0)+y2*PARAM(0)+z2)*PARAM(1))
      -exp(-(x2*PARAM(0)+y2+z2*PARAM(0))*PARAM(1)) -exp(-(x2+y2*PARAM(0)+z2*PARAM(0))*PARAM(1));
    return(-r*PARAM(2));
}

DBL f_spikes_2d(FPUContext *ctx, DBL *ptr, unsigned int) // 63
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[63].parameter_cnt == 7u);
    /*  2-D distribution */
    return(-( PARAM(0)* cos(PARAM(1)*PARAM_X) * cos(PARAM(2)*PARAM_Z) * exp(- PARAM(3)*(PARAM_X*PARAM_X+PARAM_Z*PARAM_Z)) -PARAM_Y));
}

DBL f_spiral(FPUContext *ctx, DBL *ptr, unsigned int) // 64
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[64].parameter_cnt == 7u);
    DBL r, r2, th, temp;
    /* spiral shape                *
     *                             */
    r = sqrt(PARAM_X * PARAM_X + PARAM_Z * PARAM_Z);
    if ((PARAM_X == 0) && (PARAM_Z == 0))
        PARAM_X = 0.000001;
    th = atan2(PARAM_Z, PARAM_X);
    r = r + PARAM(0) * th / TWO_M_PI;
    r2 = fmod(r, PARAM(0)) - PARAM(0) * 0.5;

    if (PARAM(3) == 1)
        r2 = sqrt(r2 * r2 + PARAM_Y * PARAM_Y);
    else if (PARAM(3) != 0)
    {
        temp = 2 / PARAM(3);
        r2 = pow((pow(fabs(r2), temp) + pow(fabs(PARAM_Y), temp)), 1. / temp);
    }
    else
        r2 = max(fabs(r2), fabs(PARAM_Y));

    r = sqrt(PARAM_X * PARAM_X + PARAM_Y * PARAM_Y + PARAM_Z * PARAM_Z);

    return (-min(PARAM(2) - r, PARAM(1) - min(r2, r)));
}

DBL f_steiners_roman(FPUContext *ctx, DBL *ptr, unsigned int) // 65
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[65].parameter_cnt == 4u);
    DBL r, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, z2=PARAM_Z*PARAM_Z;
    r=-( x2*y2 + x2*z2 + y2*z2 + PARAM_X*PARAM_Y*PARAM_Z  );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_toupie(FPUContext *ctx, DBL *ptr, unsigned int) // 66   was f_strophoid()
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[66].parameter_cnt == 7u);
    DBL value;
    DBL e = std::max(1,(int)PARAM(3) & 0x0F); // Integer exponents 1-15
    if (PARAM(0) == 0.0)
    {
        DBL MjR = PARAM(1), MnR = PARAM(2);
        value = std::cbrt(1.0 + pow(std::hypot(PARAM_X,PARAM_Z)-MjR,e) +
                pow(PARAM_Y,2.0)) - std::cbrt(1.0 + MnR);
        return value;
    }
    else
    {
        float MjR = PARAM(1), MnR = PARAM(2);
        value = std::cbrtf(1.0 + powf(std::hypotf((float)PARAM_X,(float)PARAM_Z)-MjR,e) +
                powf((float)PARAM_Y,2.0)) - std::cbrtf(1.0 + MnR);
        return (DBL)value;
    }
}

DBL f_supertorus(FPUContext *ctx, DBL *ptr, unsigned int) // 67   was f_strophoid_2d()
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[67].parameter_cnt == 8u);
    DBL value;

    if (PARAM(0) == 0.0)
    {
        DBL MjR = PARAM(1), MnR = PARAM(2);
        DBL cp  = 2 / max(gkDBL_epsilon,fabs(PARAM(3)));
        DBL rp  = 2 / max(gkDBL_epsilon,fabs(PARAM(4)));
        value = pow(pow(fabs(pow(pow(fabs(PARAM_X),rp)
                           + pow(fabs(PARAM_Z),rp),1/rp)-MjR),cp)
                    + pow(fabs(PARAM_Y),cp),1/cp)-MnR;
        return value;
    }
    else
    {
        float MjR = PARAM(1), MnR = PARAM(2);
        float cp  = 2 / max(gkMinIsectDepthReturned,fabs(PARAM(3)));
        float rp  = 2 / max(gkMinIsectDepthReturned,fabs(PARAM(4)));
        value = powf(powf(fabs(powf(powf(fabs(PARAM_X),rp)
                           + powf(fabs(PARAM_Z),rp),1/rp)-MjR),cp)
                     + powf(fabs(PARAM_Y),cp),1/cp)-MnR;
        return (DBL)value;
    }
}

// Used to return the shape as positive values.
// Flipped in povr Apr 15, 2020 to align with standard behavior.
DBL f_superellipsoid(FPUContext *ctx, DBL *ptr, unsigned int) // 68
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[68].parameter_cnt == 6u);
    DBL value;
    if (PARAM(0) == 0.0)
    {
        DBL p = 2 / PARAM(1), n = 1 / PARAM(2);
        value = -1+pow((pow((pow(fabs(PARAM_X),p)
                            +pow(fabs(PARAM_Y),p)),PARAM(1)*n)
                            +pow(fabs(PARAM_Z),2*n)),PARAM(2)*.5);
    }
    else
    {
        float p1 = PARAM(1), p2 = PARAM(2);
        float p = 2 / p1, n = 1 / p2;
        value = -1+powf((powf((powf(fabs(PARAM_X),p)
                            +powf(fabs(PARAM_Y),p)),p1*n)
                            +powf(fabs(PARAM_Z),2*n)),p2*.5);
    }
    return value;
}

DBL f_th(FPUContext *ctx, DBL *ptr, unsigned int) // 69
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[69].parameter_cnt == 3u);
    return( atan2(PARAM_X,PARAM_Z) );
}

DBL f_torus(FPUContext *ctx, DBL *ptr, unsigned int) // 70
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[70].parameter_cnt == 5u);
    // Often see form: pow(sqrt(x*x+z*z)-0.5,2) + y*y - pow(0.05,2) With last
    // minor radius calc as a constant so fewer calculations in total. However,
    // the value gradient is much worse/higher. In POV-Ray the performance
    // therefore worse (2.8x slower).

    PARAM_X = std::hypot(PARAM_X,PARAM_Z) - PARAM(0);
    return -PARAM(1) + std::hypot(PARAM_X,PARAM_Y);
}

DBL f_repeat(FPUContext *ctx, DBL *ptr, unsigned int) // 71   was f_torus2
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[71].parameter_cnt == 16u);
    // 0=x, 1=y, 2=z, 3=cells val (0 to 1), 4=cells val (-1 to 1)
    int xyzOrc = std::min(4,(int)PARAM(0) & 0x07);

    DBL value=0.0, x=PARAM_X, y=PARAM_Y, z=PARAM_Z;

    DBL xcnt=PARAM(1), ycnt=PARAM(2), zcnt=PARAM(3);
    DBL xdlt=PARAM(4), ydlt=PARAM(5), zdlt=PARAM(6);
    DBL xdir=PARAM(7), ydir=PARAM(8), zdir=PARAM(9);

    if (xyzOrc >= 3)
    {
        DBL xcellsOff=floor(PARAM(10)), ycellsOff=floor(PARAM(11)), zcellsOff=floor(PARAM(12));
        value = min(gPatternRands(
                        Hash3d(
                            (int)floor((x + xcellsOff + gkDBL_epsilon)*(1.0/fabs(xdlt))),
                            (int)floor((y + ycellsOff + gkDBL_epsilon)*(1.0/fabs(ydlt))),
                            (int)floor((z + zcellsOff + gkDBL_epsilon)*(1.0/fabs(zdlt)))
                        )
                    ),1.0);
        if (xyzOrc == 4)
            return (value-0.5)*2.0;
        else
            return value;
    }

    auto lambdaRepeatVal = [&value]
         (const DBL xyz, const DBL cnt, const DBL dlt, const DBL dir,
          const bool possibleInRange) -> void
    {
        if (possibleInRange && (fabs(xyz)<=(cnt*dlt)))
        {
            if (dir == 0.0)
                value = fmod(fabs(xyz),dlt)-dlt/2.0;
            else if ((dir < 0.0) && (xyz < 0.0))
                value = fmod(fabs(xyz),dlt)-dlt/2.0;
            else if ((dir > 0.0) && (xyz > 0.0))
                value = fmod(xyz,dlt)-dlt/2.0;
        }
        else
        {
            value = -dlt/2.0;
        }
        // Should I do fabs of dlt and then value = fabs(dlt) - value on odd floor(xyz) ??? to flip x,y,z ?
    };

    bool validx = ((xdir == 0.0) || ((xdir < 0.0) && (x < 0.0)) || ((xdir > 0.0) && (x > 0.0))) ? true : false;
    bool validy = ((ydir == 0.0) || ((ydir < 0.0) && (y < 0.0)) || ((ydir > 0.0) && (y > 0.0))) ? true : false;
    bool validz = ((zdir == 0.0) || ((zdir < 0.0) && (z < 0.0)) || ((zdir > 0.0) && (z > 0.0))) ? true : false;

    if (xyzOrc == 0)
    {
        lambdaRepeatVal(x,xcnt,xdlt,xdir,(validy && validz));
    }
    else if (xyzOrc == 1)
    {
        lambdaRepeatVal(y,ycnt,ydlt,ydir,(validx && validz));
    }
    else if (xyzOrc == 2)
    {
        lambdaRepeatVal(z,zcnt,zdlt,zdir,(validx && validy));
    }

    return value;
}

DBL f_torus_gumdrop(FPUContext *ctx, DBL *ptr, unsigned int) // 72
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[72].parameter_cnt == 4u);
    DBL r, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, z2=PARAM_Z*PARAM_Z;
    r=-( 4*(x2*x2 + (y2 + z2)*(y2 + z2)) + 17 * x2 * (y2 + z2) - 20 * (x2 + y2 + z2) + 17);
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_umbrella(FPUContext *ctx, DBL *ptr, unsigned int) // 73
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[73].parameter_cnt == 4u);
    DBL r;
    r=-( PARAM_X*PARAM_X - PARAM_Y*PARAM_Z*PARAM_Z );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_witch_of_agnesi(FPUContext *ctx, DBL *ptr, unsigned int) // 74
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[74].parameter_cnt == 5u);
    DBL r;
    r=-( PARAM(1) * (PARAM_Y - 1) + (PARAM_X*PARAM_X + PARAM_Z*PARAM_Z) * PARAM_Y   );
    return( min(10., max(PARAM(0)*r,-10.)) );
}

DBL f_witch_of_agnesi_2d(FPUContext *ctx, DBL *ptr, unsigned int) // 75
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[75].parameter_cnt == 9u);
    DBL r, x2=PARAM_X*PARAM_X, y2=PARAM_Y*PARAM_Y, th;
    ROT2D(PARAM(3),PARAM(4),PARAM(5))
    r=-( PARAM(1)*PARAM(1) * PARAM_Y + x2 * PARAM_Y - PARAM(2)  );
    return( min(10.0, max(PARAM(0)*r,-10.0)) );
}

DBL f_noise3d(FPUContext *ctx, DBL *ptr, unsigned int) // 76
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[76].parameter_cnt == 4u);
    Vector3d Vec = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    int ngen = (int)PARAM(0) & 0x03; // Force explicitness. 1-3 Valid currently 2-usual use.
    return Noise(Vec, ngen);
}

DBL f_pattern(FPUContext *ctx, DBL *ptr, unsigned int fn) // 77
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[77].parameter_cnt == 3u);
    Vector3d TPoint;
    Vector3d Vec = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    FunctionCode *f = ctx->functionvm->GetFunction(fn);

    if (f->private_data == nullptr)
        return 0.0;

    Warp_EPoint (TPoint, Vec, reinterpret_cast<const TPATTERN *>(f->private_data));

    return Evaluate_TPat(reinterpret_cast<const TPATTERN *>(f->private_data), TPoint, nullptr, nullptr, ctx->threaddata);
}

DBL f_snoise3d(FPUContext *ctx, DBL *ptr, unsigned int) // 78
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[78].parameter_cnt == 4u);
    Vector3d Vec = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    int ngen = (int)PARAM(0) & 0x03;  // Force explicitness. 1-3 Valid currently 2-usual use.
    DBL value = 2.0 * Noise(Vec, ngen) -1.0;
    return value;
}

DBL f_npmod(FPUContext *ctx, DBL *ptr, unsigned int) // 79
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[79].parameter_cnt == 3u);
    DBL value=PARAM_X, valueDenominator=PARAM_Y, valueMultiplier=PARAM_Z;

    if ((value == 0.0) || (valueDenominator <= 0.0))
    {
        return value;
    }
    else
    {
        if (fabs(value) > valueDenominator)
        {
            DBL tmpValue = value;
            DBL tmpCount = int(value / valueDenominator);
            value = fmod(tmpValue, valueDenominator);
            if (fabs(value) < gkDBL_epsilon)
            {
                return tmpValue / tmpCount * valueMultiplier;
            }
            else
            {
                return value * valueMultiplier;
            }
        }
        else
        {
            return value * valueMultiplier;
        }
    }
}

DBL f_eblob(FPUContext *ctx, DBL *ptr, unsigned int) // 80
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[80].parameter_cnt == 19u);
    const int width=8, height=4;
    DBL ary[width][height];
    for (int i=0; i<width; i++)
    {
        ary[i][0] = PARAM(i);                // incoming values
        ary[i][1] = fabs(PARAM(width+i)) / exp(1); // psuedo normalized strengths
    }

    DBL eDelta = PARAM_X;                    // offset between 1/exp()s. 0.001 good.
    int eCount = (int)PARAM_Y & 0x0F;        // count of meaningful passed values.
    bool returnAsPatternValue = (PARAM_Z == 0) ? false : true;
    const DBL limitValToExp = 38.0;          // Clamp +-value for values passed to exp().
    DBL threshold = (2.0 * eDelta) + (eDelta / 10.0);

    for (int i=0; i<width; i++)
        ary[i][2] = max(-limitValToExp,min(limitValToExp,(ary[i][0] * ary[i][1])));

    DBL value = threshold;
    for (int i=0; i<min(width,eCount); i++)
    {
        DBL valPolarity = (PARAM(width+i) >= 0.0) ? 1.0 : -1.0;
        if (ary[i][1] == 0.0)
        {
            ary[i][3] = ary[i][0];
            value = min((value * valPolarity),ary[i][0]);
        }
        else
        {
            DBL valOneOverEp = 1.0 / exp(ary[i][2] + eDelta);
            DBL valOneOverEn = 1.0 / exp(ary[i][2] - eDelta);
            ary[i][3] = valOneOverEp - valOneOverEn;
            value += ary[i][3] * valPolarity;
        }
    }

    if ((!returnAsPatternValue) || (value >= 0.0) || (eCount <= 1))
    {
        return value;
    }
    else
    {
        DBL valPattern = 0.0;
        DBL mapCnt = 0;
        switch (eCount)
        {
            case 2: mapCnt = 1+ 2; break;
            case 3: mapCnt = 1+ 3; break;
            case 4: mapCnt = 1+ 6; break;
            case 5: mapCnt = 1+10; break;
            case 6: mapCnt = 1+15; break;
            case 7: mapCnt = 1+21; break;
            case 8: mapCnt = 1+28; break;
            default: mapCnt = 1+ 2;
        }

        DBL valMostNegTwoCombo = HUGE_VAL;
        const DBL localEpsilon = 1e-6; // Align with map entry formatting to six sig digits.
        int cnt = 0;
        for (int i = 0; i < eCount; i++)
        {
            for (int j = i+1; j < eCount; j++)
            {
                cnt++;
                DBL pairVal = ary[i][3] + ary[j][3];
                if (pairVal < valMostNegTwoCombo)
                {
                    valMostNegTwoCombo = pairVal;
                    DBL valMapLo = cnt/mapCnt;
                    DBL valMapHi = (cnt+1)/mapCnt;
                    DBL valMapDelta = valMapHi - valMapLo - (2.0*localEpsilon);
                    if ((ary[i][3] < 0.0) && (ary[j][3] < 0.0))
                    {
                        if ((PARAM_Z == 1.1) && (fabs(ary[i][3]) > threshold))
                        {
                            valPattern = valMapLo + localEpsilon;
                        }
                        else if ((PARAM_Z == 1.1) && (fabs(ary[j][3]) > threshold))
                        {
                            valPattern = valMapHi - localEpsilon;
                        }
                        else
                        {
                            DBL valHiWeightLo = ary[i][3]/(ary[i][3]+ary[j][3]);
                            DBL valHiWeightHi = ary[j][3]/(ary[i][3]+ary[j][3]);
                            if ((PARAM_Z > 1.1) && (PARAM_Z <= 6.0))
                            {   // poly wave / pow() to speed change from middle to ends.
                                DBL valPow = PARAM_Z;
                                valPattern = valMapLo + valMapDelta*0.5; // middle
                                if (valHiWeightLo < valHiWeightHi)
                                {
                                    valPattern += ((valMapDelta * 0.5) * (1.0 - pow(valHiWeightLo/0.5,valPow)));
                                    valPattern -= localEpsilon;
                                }
                                else
                                {
                                    valPattern -= ((valMapDelta * 0.5) * (1.0 - pow(valHiWeightHi/0.5,valPow)));
                                    valPattern += localEpsilon;
                                }
                            }
                            else
                            {   // Linear interpolation between two contributing values.
                                if (valHiWeightLo < valHiWeightHi)
                                {
                                    valPattern = valMapLo + localEpsilon + (valMapDelta * valHiWeightHi);
                                }
                                else
                                {
                                    valPattern = valMapHi - localEpsilon - (valMapDelta * valHiWeightLo);
                                }
                            }
                        }
                    }
                    else if (ary[i][3] < 0.0)
                    {
                        valPattern = valMapLo + localEpsilon;
                    }
                    else if (ary[j][3] < 0.0)
                    {
                        valPattern = valMapHi - localEpsilon;
                    }
                    else
                    {
                        valPattern = 0.0;
                    }
                }
            }
        }
        return valPattern;
    }
}

void f_pigment(FPUContext *ctx, DBL *ptr, unsigned int fn, unsigned int sp) // 0
{
    POV_STATIC_ASSERT(POVFPU_TrapSTable[0].parameter_cnt == 3u);
    Vector3d Vec = Vector3d(PARAM_N_X(5), PARAM_N_Y(5), PARAM_N_Z(5));
    TransColour Col;
    RGBFTColour rgbftCol;
    FunctionCode *f = ctx->functionvm->GetFunction(fn);

    if (f->private_data == nullptr)
    {
        ctx->SetLocal(sp + pRED,    0.0);
        ctx->SetLocal(sp + pGREEN,  0.0);
        ctx->SetLocal(sp + pBLUE,   0.0);
        ctx->SetLocal(sp + pFILTER, 0.0);
        ctx->SetLocal(sp + pTRANSM, 0.0);
        return;
    }

    Compute_Pigment(Col, reinterpret_cast<const PIGMENT *>(f->private_data), Vec, nullptr, nullptr, ctx->threaddata);
    rgbftCol = ToRGBFTColour(Col);

    ctx->SetLocal(sp + pRED,    rgbftCol.red());
    ctx->SetLocal(sp + pGREEN,  rgbftCol.green());
    ctx->SetLocal(sp + pBLUE,   rgbftCol.blue());
    ctx->SetLocal(sp + pFILTER, rgbftCol.filter());
    ctx->SetLocal(sp + pTRANSM, rgbftCol.transm());
}

void f_transform(FPUContext *ctx, DBL *ptr, unsigned int fn, unsigned int sp) // 1
{
    POV_STATIC_ASSERT(POVFPU_TrapSTable[1].parameter_cnt == 3u);
    Vector3d Vec = Vector3d(PARAM_N_X(3), PARAM_N_Y(3), PARAM_N_Z(3));
    Vector3d Result;
    FunctionCode *f = ctx->functionvm->GetFunction(fn);

    if (f->private_data == nullptr)
    {
        ctx->SetLocal(sp + X, 0.0);
        ctx->SetLocal(sp + Y, 0.0);
        ctx->SetLocal(sp + Z, 0.0);
        return;
    }

    MTransPoint(Result, Vec, reinterpret_cast<const TRANSFORM *>(f->private_data));

    ctx->SetLocal(sp + X, Result[X]);
    ctx->SetLocal(sp + Y, Result[Y]);
    ctx->SetLocal(sp + Z, Result[Z]);
}

void f_spline(FPUContext *ctx, DBL *ptr, unsigned int fn, unsigned int sp) // 2
{
    POV_STATIC_ASSERT(POVFPU_TrapSTable[2].parameter_cnt == 1u);
    EXPRESS Result;
    FunctionCode *f = ctx->functionvm->GetFunction(fn);
    int Terms;

    if (f->private_data == nullptr)
    {
        ctx->SetLocal(sp + X, 0.0);
        ctx->SetLocal(sp + Y, 0.0);
        ctx->SetLocal(sp + Z, 0.0);
        return;
    }

    Terms = (reinterpret_cast<GenericSpline *>(f->private_data))->Terms;

    Get_Spline_Val(reinterpret_cast<GenericSpline *>(f->private_data), PARAM_N_X(Terms), Result, &Terms);

    ctx->SetLocal(sp + X, Result[X]);
    ctx->SetLocal(sp + Y, Result[Y]);
    if(Terms > 2)
    {
        ctx->SetLocal(sp + Z, Result[Z]);
        if(Terms > 3)
        {
            ctx->SetLocal(sp + T, Result[T]);
            if(Terms > 4)
                ctx->SetLocal(sp + T + 1, Result[T + 1]);
        }
    }
}

void f_patmods(FPUContext *ctx, DBL *ptr, unsigned int fn, unsigned int sp) // 3
{
    POV_STATIC_ASSERT(POVFPU_TrapSTable[3].parameter_cnt == 3u);
    Vector3d TPoint;
    Vector3d Vec = Vector3d(PARAM_N_X(3), PARAM_N_Y(3), PARAM_N_Z(3));
    FunctionCode *f = ctx->functionvm->GetFunction(fn);

    if (f->private_data == nullptr)
    {
        ctx->SetLocal(sp + X, 0.0);
        ctx->SetLocal(sp + Y, 0.0);
        ctx->SetLocal(sp + Z, 0.0);
        return;
    }

    Warp_EPoint (TPoint, Vec, reinterpret_cast<const TPATTERN *>(f->private_data));

    ctx->SetLocal(sp + X, TPoint.x());
    ctx->SetLocal(sp + Y, TPoint.y());
    ctx->SetLocal(sp + Z, TPoint.z());
    return;
}


}
// end of namespace pov
