//******************************************************************************
///
/// @file vfe/unix/unixconsole.cpp
///
/// Adapted from @ref vfe/win/console/winconsole.cpp
///
/// @copyright
/// @parblock
///
/// Persistence of Vision Ray Tracer ('POV-Ray') version 3.8.
/// Copyright 1991-2020 Persistence of Vision Raytracer Pty. Ltd.
///
/// POV-Ray is free software: you can redistribute it and/or modify
/// it under the terms of the GNU Affero General Public License as
/// published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
///
/// POV-Ray is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Affero General Public License for more details.
///
/// You should have received a copy of the GNU Affero General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.
///
/// ----------------------------------------------------------------------------
///
/// POV-Ray is based on the popular DKB raytracer version 2.12.
/// DKBTrace was originally written by David K. Buck.
/// DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
///
/// @endparblock
///
//******************************************************************************

// C++ variants of C standard header files
#include <csignal>
#include <cstdlib>

// C++ standard header files
#include <memory>
#include <mutex>
#include <string>
#include <thread>

// Boost library files
#include <boost/algorithm/string.hpp>

// Other library header files
#include <termios.h>
#include <unistd.h>
#include <sys/select.h>

// from directory "vfe"
#include "vfe.h"

// from directory "unix"
#include "disp.h"
#include "disp_text.h"
#include "disp_sdl.h"
#include "disp_sdl2.h"
#include "disp_x11.h"

// from directory "source"
#include "backend/povray.h"

namespace pov_frontend
{
    std::shared_ptr<Display> gDisplay;
}
// end of namespace pov_frontend

using namespace vfe;
using namespace vfePlatform;

enum FlagUnixConsole
{
    kFlagHelp,
    kFlagVersion,
    kFlagGeneration,
    kFlagLibraries,
    kFlagLicense,
    kFlagAuthors
};

enum DispMode
{
    DISP_MODE_NONE,
    DISP_MODE_TEXT,
    DISP_MODE_X11,
    DISP_MODE_SDL,
    DISP_MODE_SDL2
};

static DispMode gDisplayMode;

enum ReturnValue
{
    RETURN_OK=0,
    RETURN_ERROR,
    RETURN_USER_ABORT
};

static bool gCancelRender = false;

// for handling asynchronous (external) signals
static int gSignalNumber = 0;
static std::mutex gSignalMutex;
static volatile bool gTerminateSignalHandler = false;


// TODO FIXME - This way to handle signals seems rather wonky, as it is subject
// to a potential race condition on `gSignalNumber` between `ProcessSignal()`
// polling the signal and `SignalHandler()` processing the next signal.
// Wouldn't it be smarter to register proper signal handlers?
//
static void SignalHandler (void)
{
    sigset_t  sigset;
    timespec  timeout;
    int       signum;

    sigfillset(&sigset);
    timeout.tv_sec  = 0;
    timeout.tv_nsec = 10000000; // 10 ms

    while(!gTerminateSignalHandler)
    {
        // Wait till a signal is caught.
#ifdef HAVE_SIGTIMEDWAIT
        signum = sigtimedwait(&sigset, nullptr, &timeout);
        if (signum == -1)
            continue; // Error. Presumably timed out; check whether to end the task.
#else
        (void)sigwait(&sigset, &signum);
#endif

        // Got a signal.
        std::lock_guard<std::mutex> lock(gSignalMutex);
        gSignalNumber = signum;
    }
}


static void ProcessSignal (void)
{
    std::lock_guard<std::mutex> lock(gSignalMutex);

    switch (gSignalNumber)
    {
        case  0:
            break;
#ifdef SIGQUIT
        case SIGQUIT:
            fprintf(stderr, "\n%s: received signal SIGQUIT: Quit; requested render cancel\n", PACKAGE);
            gCancelRender = true;
            break;
#endif
#ifdef SIGTERM
        case SIGTERM:
            fprintf(stderr, "\n%s: received signal SIGTERM: Termination; requested render cancel\n", PACKAGE);
            gCancelRender = true;
            break;
#endif
#ifdef SIGINT
        case SIGINT:
            fprintf(stderr, "\n%s: received signal SIGINT: Interrupt; requested render cancel\n", PACKAGE);
            gCancelRender = true;
            break;
#endif
#ifdef SIGPIPE
        case SIGPIPE:
            fprintf(stderr, "\n%s: received signal SIGPIPE: Broken pipe; requested render cancel\n", PACKAGE);
            gCancelRender = true;
            break;
#endif
        case SIGCHLD:
            // for now, ignore this (side-effect of the shell-out code).
            // once properly implemented, the shell-out code would want to know this has happened, though.
            break;

        default:
            // fprintf(stderr, "\n%s: received signal %d\n", PACKAGE, gSignalNumber);
            break;
    }
    gSignalNumber = 0;
}

static vfeDisplay *UnixDisplayCreator (unsigned int width, unsigned int height, vfeSession *session, bool visible)
{
    UnixDisplay *display = GetRenderWindow () ;
    switch (gDisplayMode)
    {
#ifndef X_DISPLAY_MISSING
        case DISP_MODE_X11:
            #if POV_MULTITHREADED
            XInitThreads(); // In practice this seems to fix
                            // "[xcb] Most likely this is a multi-threaded client and XInitThreads has not been called"
                            // errors during animation renders. https://www.x.org/archive/X11R6.8.2/doc/XInitThreads.3.html
            #endif

            if (display != nullptr && display->GetWidth() == width && display->GetHeight() == height)
            {
                UnixDisplay *p = new UnixX11Display (width, height, session, false) ;
                if (p->TakeOver (display))
                    return p;
                delete p;
            }
            return new UnixX11Display (width, height, session, visible) ;
            break;
#ifdef HAVE_LIBSDL2
        case DISP_MODE_SDL2:
        case DISP_MODE_SDL:
            if (display != nullptr && display->GetWidth() == width && display->GetHeight() == height)
            {
                UnixDisplay *p = new UnixSDL2Display (width, height, session, false) ;
                if (p->TakeOver (display))
                    return p;
                delete p;
            }
            return new UnixSDL2Display (width, height, session, visible) ;
            break;
#endif
#ifdef HAVE_LIBSDL
        case DISP_MODE_SDL2:
        case DISP_MODE_SDL:
            if (display != nullptr && display->GetWidth() == width && display->GetHeight() == height)
            {
                UnixDisplay *p = new UnixSDLDisplay (width, height, session, false) ;
                if (p->TakeOver (display))
                    return p;
                delete p;
            }
            return new UnixSDLDisplay (width, height, session, visible) ;
            break;
#endif
#endif
        case DISP_MODE_TEXT:
            return new UnixTextDisplay (width, height, session, visible) ;
            break;
        default:
            return nullptr;
    }
}


static void PrintStatus (vfeSession *session)
{
    // TODO -- when invoked while processing "--help" command-line switch,
    //         GNU/Linux customs would be to print to stdout (among other differences).

    std::string str;
    vfeSession::MessageType type;
    static vfeSession::MessageType lastType = vfeSession::mUnclassified;

    while (session->GetNextCombinedMessage (type, str))
    {
        if (type != vfeSession::mGenericStatus)
        {
            if (lastType == vfeSession::mGenericStatus)
                fprintf (stderr, "\n") ;
            fprintf (stderr, "%s\n", str.c_str());
        }
        else
            fprintf (stderr, "%s\r", str.c_str());
        lastType = type;
    }
}

static void PrintStatusChanged (vfeSession *session, State force = kUnknown)
{
    if (force == kUnknown)
        force = session->GetBackendState();
    switch (force)
    {
        case kParsing:
            fprintf (stderr, "==== [Parsing...] ==========================================================\n");
            break;
        case kRendering:
#ifndef X_DISPLAY_MISSING
            if ((gDisplay != nullptr) &&
                (
                 (gDisplayMode == DISP_MODE_X11)
                || (gDisplayMode == DISP_MODE_SDL)
                || (gDisplayMode == DISP_MODE_SDL2)
                ))
            {
                fprintf (stderr, "==== [Rendering... Press p to pause, q to quit] ============================\n");
            }
            else
            {
                fprintf (stderr, "==== [Rendering...] ========================================================\n");
            }
#else
            fprintf (stderr, "==== [Rendering...] ========================================================\n");
#endif
            break;
        case kPausedRendering:
#ifndef X_DISPLAY_MISSING
            if ((gDisplay != nullptr) &&
                (
                 (gDisplayMode == DISP_MODE_X11)
                || (gDisplayMode == DISP_MODE_SDL)
                || (gDisplayMode == DISP_MODE_SDL2)
                ))
            {
                fprintf (stderr, "==== [Paused... Press p to resume] =========================================\n");
            }
            else
            {
                fprintf (stderr, "==== [Paused...] ===========================================================\n");
            }
#else
            fprintf (stderr, "==== [Paused...] ===========================================================\n");
#endif
            break;
        default:
            // Do nothing special.
            break;
    }
}


static void PrintAuthors(void)
{
    fprintf(stderr,
        "=======\n%s %s \n\n",
        PACKAGE_NAME, POV_RAY_VERSION
    );
    fprintf(stderr,
        "Primary Architects/Developers: (Alphabetically)\n"
        "  Chris Cason         Thorsten Froehlich  Christoph Lipka\n"
        "\n"
        "With Assistance From: (Alphabetically)\n"
        "  Nicolas Calimet     Jerome Grimbert     James Holsenback\n"
        "  Christoph Hormann   Nathan Kopp         Juha Nieminen\n"
        "  William F. Pokorny\n"
        "\n"
        "Past Contributors: (Alphabetically)\n"
        "  Steve Anger         Eric Barish         Dieter Bayer\n"
        "  David K. Buck       Nicolas Calimet     Chris Cason\n"
        "  Aaron A. Collins    Chris Dailey        Steve Demlow\n"
        "  Andreas Dilger      Alexander Enzmann   Dan Farmer\n"
        "  Thorsten Froehlich  Mark Gordon         Jerome Grimbert\n"
        "  James Holsenback    Christoph Hormann   Mike Hough\n"
        "  Chris Huff          Kari Kivisalo       Nathan Kopp\n"
        "  Lutz Kretzschmar    Christoph Lipka     Jochen Lippert\n"
        "  Pascal Massimino    Jim McElhiney       Douglas Muir\n"
        "  Juha Nieminen       Ron Parker          William F. Pokorny\n"
        "  Bill Pulver         Eduard Schwan       Wlodzimierz Skiba\n"
        "  Robert Skinner      Yvo Smellenbergh    Zsolt Szalavari\n"
        "  Scott Taylor        Massimo Valentini   Timothy Wegner\n"
        "  Drew Wells          Chris Young\n"
        "\n"
        "Other contributors are listed in the documentation at:\n"
        "  http://wiki.povray.org/content/Documentation\n"
        "    and at:\n"
        "  https://github.com/POV-Ray/povray/graphs/contributors\n"
        "\n"
        "POV-Ray is based on DKBTrace 2.12 by David K. Buck and\n"
        "Aaron A. Collins.\n"
    );
    fprintf(stderr,
        "\n%s\n\n",
        POV_RAY_COPYRIGHT
    );
}


static void PrintVersion(void)
{
    // TODO -- GNU/Linux customs would be to print to stdout (among other differences).
    //
    // At the moment keeping aligned with povray.cpp BuildInitInfo
    fprintf(stderr,
        "Persistence of Vision(tm) Ray Tracer    (see --license)\n"
        "%s\n"
        "%s %s\n"
        "%s%s\n\n",
        POV_RAY_COPYRIGHT,
        PACKAGE_NAME, POV_RAY_VERSION,
        DISTRIBUTION_MESSAGE_1, DISTRIBUTION_MESSAGE_2
    );
    fprintf(stderr,
        "Built-in features:\n"
        "  Supported image formats:   %s\n"
        "  Unsupported image formats: %s\n\n",
        BUILTIN_IMG_FORMATS, MISSING_IMG_FORMATS
    );
//      "  I/O restrictions:          %s\n"
//      "  X Window display:          %s\n"
//      BUILTIN_IO_RESTRICTIONS, BUILTIN_XWIN_DISPLAY,

    fprintf(stderr,
        "Compilation settings:\n"
        "  Build architecture:  %s\n"
        "  Built/Optimized for: %s\n"
        "  Compiler vendor:     %s\n"
        "  Compiler version:    %s\n"
        "  Compiler flags:      %s\n",
        BUILD_ARCH, BUILT_FOR, COMPILER_VENDOR, COMPILER_VERSION, CXXFLAGS
    );
}

//------------------------------------- Libraries
#include <boost/version.hpp>
#ifndef LIBJPEG_MISSING
    #include <jpeglib.h>
#endif
#ifndef OPENEXR_MISSING
    #include <IlmBaseConfig.h>
    #include <OpenEXRConfig.h>
#endif
#ifndef LIBPNG_MISSING
    #include <png.h>
#endif
#ifndef LIBTIFF_MISSING
    #include <tiffvers.h>
#endif
#ifndef LIBZ_MISSING
    #include <zlib.h>
#endif

static void PrintLibraries(void)
{
    char buffer[200];
  //ExtractLibraryVersion(zlibVersion(), buffer);

    fprintf(stderr,
        "\nBoost %d.%d, http://www.boost.org/\n",
        BOOST_VERSION/100000, BOOST_VERSION/100%1000
    );

#ifndef LIBJPEG_MISSING
    fprintf(stderr,
        "\nlibjpeg %d.%d, Copyright (C) 1991-1998, Thomas G. Lane.\n"
        "Modified 2002-2009 by Guido Vollbeding.\n",
        JPEG_LIB_VERSION/10, JPEG_LIB_VERSION%10
    );
#endif

#ifndef OPENEXR_MISSING
    fprintf(stderr,
         "\n%s and %s,\n"
         "Copyright (c) 2002-2011 Industrial Light & Magic.\n",
         OPENEXR_PACKAGE_STRING, ILMBASE_PACKAGE_STRING
    );
#endif

#ifndef LIBPNG_MISSING
    fprintf(stderr,
        "%s\n",png_get_copyright(nullptr)
    );
#endif

#ifndef LIBTIFF_MISSING
    fprintf(stderr,
        "%s\n",TIFFLIB_VERSION_STR
    );
#endif

#ifndef LIBZ_MISSING
    fprintf(stderr,
        "\nzlib %s, Copyright (C) 1995-2017 Jean-loup Gailly and Mark Adler\n",
        ZLIB_VERSION
    );
#endif
    fprintf(stderr,"\n");
}


static void PrintLicense(void)
{
    fprintf(stderr,
        "\n"
        "POV-Ray is free software: you can redistribute it and/or modify\n"
        "it under the terms of the GNU Affero General Public License as\n"
        "published by the Free Software Foundation, either version 3 of the\n"
        "License, or (at your option) any later version.\n"
        "\n"
        "POV-Ray is distributed in the hope that it will be useful,\n"
        "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
        "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
        "GNU Affero General Public License for more details.\n"
        "\n"
        "You should have received a copy of the GNU Affero General Public License\n"
        "along with this program.  If not, see <http://www.gnu.org/licenses/>.\n"
        "\n"
        "All of the above applies to the unoffical 'povr' branch of POV-Ray.\n"
        "\n"
    );
}


static void PrintGeneration(void)
{
    fprintf(stdout, "%s\n", POV_RAY_GENERATION POV_RAY_BETA_SUFFIX);
}


static void ErrorExit(vfeSession *session)
{
    fprintf(stderr, "%s\n", session->GetErrorString());
    session->Shutdown();
    delete session;
    std::exit(RETURN_ERROR);
}


static void CancelRender(vfeSession *session)
{
    session->CancelRender();  // request the backend to cancel
    PrintStatus (session);
    while (session->GetBackendState() != kReady)  // wait for the render to effectively shut down
        Delay(10);
    PrintStatus (session);
}


static void PauseWhenDone(vfeSession *session)
{
    GetRenderWindow()->UpdateScreen(true);
    GetRenderWindow()->PauseWhenDoneNotifyStart();
    while (GetRenderWindow()->PauseWhenDoneResumeIsRequested() == false)
    {
        ProcessSignal();
        if (gCancelRender)
            break;
        else
            Delay(10);
    }
    GetRenderWindow()->PauseWhenDoneNotifyEnd();
}


static void TerminateSignalHandler(std::thread* sigthread)
{
    gTerminateSignalHandler = true;
#ifdef HAVE_SIGTIMEDWAIT
    // `std::thread`s must be `join`ed or `detach`ed before destruction,
    // otherwise their destructor causes ungraceful termination of the entire
    // program.
    sigthread->join();
#else
    // `std::thread`s must be `join`ed or `detach`ed before destruction,
    // otherwise their destructor causes ungraceful termination of the entire
    // program. `join` is not an option because we can't reliably make the
    // thread terminate itself, so we must trust the OS that it will kill any
    // `detach`ed threads when their parent process exits.
    // TODO - This is a hackish solution.
    sigthread->detach();
#endif
}

int main (int argc, char **argv)
{
    vfeUnixSession   *session;
    vfeStatusFlags    flags;
    vfeRenderOptions  opts;
    ReturnValue       retval = RETURN_OK;
    sigset_t          sigset;
    std::thread*      sigthread;
    char **           argv_copy=argv; /* because argv is updated later */
    int               argc_copy=argc; /* because it might also be updated */

    /*fprintf(stderr, "%s: This is a RELEASE CANDIDATE version of POV-Ray. General distribution is discouraged.\n", PACKAGE);*/

    // block some signals for this thread as well as those created afterwards
    sigemptyset(&sigset);

#ifdef SIGQUIT
    sigaddset(&sigset, SIGQUIT);
#endif
#ifdef SIGTERM
    sigaddset(&sigset, SIGTERM);
#endif
#ifdef SIGINT
    sigaddset(&sigset, SIGINT);
#endif
#ifdef SIGPIPE
    sigaddset(&sigset, SIGPIPE);
#endif
#ifdef SIGCHLD
    sigaddset(&sigset, SIGCHLD);
#endif

    pthread_sigmask(SIG_BLOCK, &sigset, nullptr);

    // create the signal handling thread
    sigthread = new std::thread(SignalHandler);

    session = new vfeUnixSession();
    if (session->Initialize(nullptr, nullptr) != vfeNoError)
        ErrorExit(session);

    // display mode registration
#ifndef X_DISPLAY_MISSING
    UnixX11Display::Register(session);
#ifdef HAVE_LIBSDL2
    UnixSDL2Display::Register(session);
#endif
#ifdef HAVE_LIBSDL
    UnixSDLDisplay::Register(session);
#endif
#endif
    UnixTextDisplay::Register(session);

    // default number of work threads: number of CPUs or 4
    int nthreads = 1;
#ifdef _SC_NPROCESSORS_ONLN  // online processors
    nthreads = sysconf(_SC_NPROCESSORS_ONLN);
#endif
#ifdef _SC_NPROCESSORS_CONF  // configured processors
    if (nthreads < 2)
        nthreads = sysconf(_SC_NPROCESSORS_CONF);
#endif
    if (nthreads < 2)
        nthreads = 4;
    opts.SetThreadCount(nthreads);

    // process command-line options
    session->GetUnixOptions()->ProcessOptions(&argc, &argv);
    {
      /* tricky part, -1 is choose the top available (oh, the personal taste...)
       * for more direct choice, degrade to DISP_MODE_NODE when not available
       */
      std::string choice = session->GetUnixOptions()->QueryOptionString("display","window");
#ifndef X_DISPLAY_MISSING
      if (choice == "x11")
      {
          gDisplayMode = DISP_MODE_X11;
      }
      else if (choice == "sdl")
      {
#ifdef HAVE_LIBSDL2
          gDisplayMode = DISP_MODE_SDL2;
#else
#ifdef HAVE_LIBSDL
          gDisplayMode = DISP_MODE_SDL;
#else
          gDisplayMode = DISP_MODE_TEXT;
#endif
#endif
      }
      else
#endif
      if (choice == "text")
      {
          gDisplayMode = DISP_MODE_TEXT;
      }
      else
      {
#ifdef HAVE_LIBSDL2
          gDisplayMode = DISP_MODE_SDL2;
#else
#ifdef HAVE_LIBSDL
          gDisplayMode = DISP_MODE_SDL;
#else
#ifndef X_DISPLAY_MISSING
          gDisplayMode = DISP_MODE_X11;
#else
          gDisplayMode = DISP_MODE_TEXT;
#endif
#endif
#endif
      }
    }

    { // Handle unix console specific flags in this block
        bool flagSet  = false;
        int  flagType = -1;
        if (session->GetUnixOptions()->isOptionSet("general", "help"))
        {
            flagSet = true;
            flagType = kFlagHelp;
        }
        else if (session->GetUnixOptions()->isOptionSet("general", "version"))
        {
            flagSet = true;
            flagType = kFlagVersion;
        }
        else if (session->GetUnixOptions()->isOptionSet("general", "generation"))
        {
            flagSet = true;
            flagType = kFlagGeneration;
        }
        else if (session->GetUnixOptions()->isOptionSet("general", "libraries"))
        {
            flagSet = true;
            flagType = kFlagLibraries;
        }
        else if (session->GetUnixOptions()->isOptionSet("general", "license"))
        {
            flagSet = true;
            flagType = kFlagLicense;
        }
        else if (session->GetUnixOptions()->isOptionSet("general", "authors"))
        {
            flagSet = true;
            flagType = kFlagAuthors;
        }
        //---
        if (flagSet)
        {
            session->Shutdown();
            switch(flagType)
            {
                case kFlagHelp:
                    PrintStatus(session);
                    session->GetUnixOptions()->PrintOptions();
                    break;
                case kFlagVersion:
            PrintVersion();
                    break;
                case kFlagGeneration:
            PrintGeneration();
                    break;
                case kFlagLibraries:
            PrintLibraries();
                    break;
                case kFlagLicense:
            PrintLicense();
                    break;
                case kFlagAuthors:
            PrintAuthors();
                    break;
                default:
                    throw POV_EXCEPTION_STRING("Internal Error. Reached unknown unix console switch.");
                    break;
            }
            TerminateSignalHandler(sigthread);
            delete sigthread;
            delete session;
            return RETURN_OK;
        }
    }


    // process settings
    char *s = std::getenv ("POVINC");
    session->SetDisplayCreator(UnixDisplayCreator);
    if (s != nullptr)                               // Must proceed Process_povray_ini defaulting to avoid picking
        opts.AddLibraryPath (s);                    // up the wrong (not POVINC) include file from a system install.
    session->GetUnixOptions()->Process_povray_ini(opts);
    while (*++argv)
        opts.AddCommand (*argv);

    // display all queued messages in console queue, allow to keep future error message at the end of the flow
    // otherwise the error get displayed before the copyright and in verbose mode the scroll
    // is big enough to loose the interesting part (and nobody looks back that much for error reports)
    PrintStatus(session);
    // set all options and start rendering
    if (session->SetOptions(opts) != vfeNoError)
    {
        // display error line and other details
        PrintStatus(session);
        // the last line of PrintStatus might ends with \r, so print an extra \n to have a clear empty line
        fprintf(stderr,"\n\nProblem with option setting\n");
        for(int loony=0;loony<argc_copy;loony++)
        {
            fprintf(stderr,"%s%c",argv_copy[loony],loony+1<argc_copy?' ':'\n');
        }
        ErrorExit(session);
    }
    if (session->StartRender() != vfeNoError)
        ErrorExit(session);

    // set inter-frame pause for animation
    if (session->RenderingAnimation() && session->GetBoolOption("Pause_When_Done", false))
        session->PauseWhenDone(true);

    // main render loop
    session->SetEventMask(stBackendStateChanged);  // immediately notify this event
    while (((flags = session->GetStatus(true, 200)) & stRenderShutdown) == 0)
    {
        ProcessSignal();
        if (gCancelRender)
        {
            CancelRender(session);
            break;
        }

        if (flags & stAnimationStatus)
            fprintf(stderr, "\nRendering frame %d of %d (#%d)\n", session->GetCurrentFrame(), session->GetTotalFrames(), session->GetCurrentFrameId());
        if (flags & stAnyMessage)
            PrintStatus (session);
        if (flags & stBackendStateChanged)
            PrintStatusChanged (session);

        if (GetRenderWindow() != nullptr)
        {
            // early exit
            if (GetRenderWindow()->HandleEvents())
            {
                gCancelRender = true;  // will set proper return value
                CancelRender(session);
                break;
            }

            GetRenderWindow()->UpdateScreen();

            // inter-frame pause
            if (session->GetCurrentFrame() < session->GetTotalFrames()
            && session->GetPauseWhenDone()
            && (flags & stAnimationFrameCompleted) != 0
            && session->Failed() == false)
            {
                PauseWhenDone(session);
                if (! gCancelRender)
                    session->Resume();
            }
        }
    }

    // pause when done for single or last frame of an animation
    if (session->Failed() == false && GetRenderWindow() != nullptr && session->GetBoolOption("Pause_When_Done", false))
    {
        PrintStatusChanged(session, kPausedRendering);
        PauseWhenDone(session);
        gCancelRender = false;
    }

    if (session->Succeeded() == false)
        retval = gCancelRender ? RETURN_USER_ABORT : RETURN_ERROR;
    session->Shutdown();
    PrintStatus (session);
    TerminateSignalHandler(sigthread);
    delete sigthread;
    delete session;

    return retval;
}
